"""Utils to manipulate dates/date times."""

import datetime
from typing import Any

import dateutil.parser

from mlcroissant._src.core.issues import Issues


def from_str_to_datetime(issues: Issues, date: str | None) -> datetime.datetime | None:
    """Converts https://schema.org/DateTime or https://schema.org/Date to datetime."""
    if date is None:
        return None
    elif isinstance(date, str):
        # https://schema.org/Date follows the ISO 8601 date format.
        # https://schema.org/DateTime has the form [-]CCYY-MM-DDThh:mm:ss[Z|(+|-)hh:mm]
        # in the ISO 8601 format.
        try:
            return dateutil.parser.isoparse(date)
        except ValueError:
            issues.add_error(
                "Dates or DateTimes should follow the [ISO 8601"
                f" format](https://en.wikipedia.org/wiki/ISO_8601). Got {date}"
            )
    else:
        issues.add_error(f"Wrong type for date. Expected str. Got {type(date)}")
    return None


def from_datetime_to_str(date: datetime.datetime | None) -> str | None:
    """Converts a datetime to https://schema.org/Date (ISO 8601 date format)."""
    if date is None:
        return None
    elif date.time() == datetime.time.min:
        return date.strftime("%Y-%m-%d")
    return date.isoformat()


def cast_date(date: Any) -> datetime.datetime | None:
    """Casts date as a datetime for any input."""
    if date is None or isinstance(date, datetime.datetime):
        return date
    elif isinstance(date, str):
        issues = Issues()
        date = from_str_to_datetime(issues, date)
        if issues.errors:
            raise ValueError(issues.errors)
        return date
    elif isinstance(date, datetime.date):
        return datetime.datetime.combine(date, datetime.time.min)
    raise ValueError(f"Wrong type for a date. Expected Date or Datetime. Got: {date}")


def cast_dates(value: Any) -> list[datetime.datetime] | None:
    """Casts dates as datetimes for any input."""
    if isinstance(value, list):
        dates = []
        for v in value:
            date = cast_date(v)
            if date is None:
                return None
            else:
                dates.append(date)
        return dates
    return cast_dates([value])



"""Tests for git."""

import pathlib
import tempfile
from unittest import mock

from etils import epath

from mlcroissant._src.core.git import download_git_lfs_file
from mlcroissant._src.core.git import is_git_lfs_file
from mlcroissant._src.core.optional import deps
from mlcroissant._src.core.path import Path

_GIT_LFS_CONTENT = lambda: """version https://git-lfs.github.com/spec/v1
oid sha256:5e2785fcd9098567a49d6e62e328923d955b307b6dcd0492f6234e96e670772a
size 309207547
"""

_NON_GIT_LFS_CONTENT = lambda: """name,age
a,1
b,2
c,3"""

_NON_ASCII_CONTENT = lambda: (255).to_bytes(1, byteorder="big")


def test_is_git_lfs_file():
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = epath.Path(tempdir)

        gitlfs_file = tempdir / "gitlfs"
        gitlfs_file.write_text(_GIT_LFS_CONTENT())
        assert is_git_lfs_file(gitlfs_file)

        no_gitlfs_file = tempdir / "no-gitlfs"
        no_gitlfs_file.write_text(_NON_GIT_LFS_CONTENT())
        assert not is_git_lfs_file(no_gitlfs_file)

        no_ascii_file = tempdir / "no-ascii"
        no_ascii_file.write_bytes(_NON_ASCII_CONTENT())
        assert not is_git_lfs_file(no_ascii_file)


def test_download_git_lfs_file():
    file = Path(
        filepath=epath.Path("/tmp/full/path.json"),
        fullpath=pathlib.PurePath("path.json"),
    )
    git = deps.git
    with mock.patch.object(git, "Git", autospec=True) as git_mock:
        download_git_lfs_file(file)
        git_mock.assert_called_once_with("/tmp/full/")
        git_mock.return_value.execute.assert_called_once_with([
            "git", "lfs", "pull", "--include", "path.json"
        ])



"""Source module."""

import enum
from typing import Any

import jsonpath_rw
from jsonpath_rw import lexer
from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.uuid import formatted_uuid_to_json
from mlcroissant._src.core.uuid import uuid_from_jsonld
from mlcroissant._src.structure_graph.base_node import Node


class FileProperty(enum.IntEnum):
    """Lists the intrinsic properties of a file that are accessible from Croissant.

    Notes:
    - Plural indicates a one-to-many relationship (one row gives many rows), while
      singular indicates a one-to-one relationship.
    - We may use camelCase to be conformed with the names in the JSON-LD Croissant
      standard.
    - We use enum.IntEnum (rather than enum.Enum) in order for FileProperty to be usable
      as column names in pd.DataFrames.

    Warning:
    - At the moment there may be an overlap with existing columns if columns
      have one of the following names.
    """

    content = 1
    filename = 2
    filepath = 3
    fullpath = 4
    lines = 5
    lineNumbers = 6


def is_file_property(file_property: str):
    """Checks if a string is a FileProperty (e.g., "content"->FileProperty.content)."""
    for possible_file_property in FileProperty:
        if possible_file_property.name == file_property:
            return True
    return False


def _cast_file_property(file_property: Any) -> FileProperty | None:
    if file_property is None:
        return None
    elif isinstance(file_property, str) and is_file_property(file_property):
        return FileProperty[file_property]
    raise ValueError(
        "Property fileProperty can only have values in `fullpath`, `filepath` and"
        f" `content`. Got: {file_property}"
    )


@mlc_dataclasses.dataclass
class Extract(Node):
    """Container for possible ways of extracting the data.

    Args:
        column: The column in a columnar format (e.g., CSV).
        file_property: The property of a file to extract.
        json_path: The JSON path if the source is a JSON.
    """

    JSONLD_TYPE = None

    column: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        exclusive_with=["file_property", "json_path"],
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_COLUMN,
    )
    file_property: FileProperty | None = mlc_dataclasses.jsonld_field(
        cast_fn=_cast_file_property,
        default=None,
        input_types=[SDO.Text],
        to_jsonld=lambda ctx, fp: fp.name if isinstance(fp, FileProperty) else fp,
        url=constants.ML_COMMONS_FILE_PROPERTY,
    )
    json_path: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_JSON_PATH,
    )


@mlc_dataclasses.dataclass
class Transform(Node):
    """Container for transformation.

    Args:
        format: The format for a date (e.g. "%Y-%m-%d %H:%M:%S.%f") or for a bounding
            box (e.g., "XYXY").
        regex: A regex pattern with a capturing group to extract information in a
            string.
        replace: A replace pattern, e.g. "pattern_to_remove/pattern_to_add".
        separator: A separator in a string to yield a list.
    """

    JSONLD_TYPE = None

    format: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        exclusive_with=["json_path", "regex", "replace", "separator"],
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_FORMAT,
    )
    json_path: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_JSON_PATH,
    )
    regex: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_REGEX,
    )
    replace: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_REPLACE,
    )
    separator: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_SEPARATOR,
    )


@mlc_dataclasses.dataclass
class Source(Node):
    r"""Python representation of sources and references.

    Croissant accepts several manners to declare sources:

    When the origin is a field:

    ```json
    "source": {
        "field": {"@id": "record_set/name"},
    }
    ```

    When the origin is a FileSet or a FileObject:

    ```json
    "source": {
        "distribution": {"@id": "my-csv"},
        "extract": {
            "column": "my-csv-column"
        }
    }
    ```

    See the specs for all supported parameters by `extract`.

    You can also add one or more transformations with `transform`:

    ```json
    "source": {
        "field": {"@id": "record_set/name"},
        "transform": {
            "format": "yyyy-MM-dd HH:mm:ss.S",
            "regex": "([^\\/]*)\\.jpg",
            "separator": "|"
        }
    }
    ```
    """

    JSONLD_TYPE = None

    distribution: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="",
        from_jsonld=lambda ctx, jsonld: uuid_from_jsonld(jsonld),
        to_jsonld=formatted_uuid_to_json,
        url=constants.SCHEMA_ORG_DISTRIBUTION,
        versions=[CroissantVersion.V_0_8],
    )
    field: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="",
        exclusive_with=["file_object", "file_set", "distribution"],
        from_jsonld=lambda ctx, jsonld: uuid_from_jsonld(jsonld),
        to_jsonld=formatted_uuid_to_json,
        url=constants.ML_COMMONS_FIELD,
    )
    file_object: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="",
        from_jsonld=lambda ctx, jsonld: uuid_from_jsonld(jsonld),
        to_jsonld=formatted_uuid_to_json,
        url=constants.ML_COMMONS_FILE_OBJECT,
    )
    file_set: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="",
        from_jsonld=lambda ctx, jsonld: uuid_from_jsonld(jsonld),
        to_jsonld=formatted_uuid_to_json,
        url=constants.ML_COMMONS_FILE_SET,
    )
    extract: Extract = mlc_dataclasses.jsonld_field(
        default_factory=Extract,
        description="",
        input_types=[Extract],
        url=constants.ML_COMMONS_EXTRACT,
    )
    transforms: list[Transform] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description="",
        input_types=[Transform],
        url=constants.ML_COMMONS_TRANSFORM,
    )

    def __bool__(self):
        """Allows to write `if not node.source` / `if node.source`."""
        return self.uuid is not None

    # TODO(https://github.com/mlcommons/croissant/issues/591): fix typing by having
    # `target_uuid` instead of overriding `Node.uuid`.
    @property
    def uuid(self) -> str | None:  # type: ignore
        """Unique identifier for the source."""
        if self.ctx.is_v0():
            return self.field or self.distribution
        else:
            return self.field or self.file_object or self.file_set

    def get_column(self) -> str | FileProperty:
        """Retrieves the name of the column associated to the source."""
        if self.uuid is None:
            raise ValueError(
                "No UUID! This case already rose an issue and should not happen at run"
                " time."
            )
        if self.extract.column:
            return self.extract.column
        elif self.extract.file_property:
            return self.extract.file_property
        elif self.extract.json_path:
            return self.extract.json_path
        else:
            if self.ctx.is_v0():
                return self.uuid.split("/")[-1]
            return self.uuid

    def check_source(self, add_error: Any):
        """Checks if the source is valid and adds error otherwise."""
        if self.extract.json_path:
            try:
                jsonpath_rw.parse(self.extract.json_path)
            except lexer.JsonPathLexerError as exception:
                add_error(
                    "Wrong JSONPath (https://goessner.net/articles/JsonPath/):"
                    f" {exception}"
                )



"""Module to manipulate UUID."""

from typing import Any
import uuid

from mlcroissant._src.core.constants import BASE_IRI
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.types import Json


def generate_uuid() -> str:
    """Generates a UUID of version 4 because it's random and simple."""
    return f"_:{str(uuid.uuid4())}"


def uuid_from_jsonld(jsonld: Json | None) -> str:
    """Retrieves uuid from a JSON-LD fragment. If no uuid, it will generate one."""
    if isinstance(jsonld, dict):
        uuid = jsonld.get("@id")
        return uuid_from_jsonld(uuid)
    elif isinstance(jsonld, list):
        return [uuid_from_jsonld(uuid) for uuid in jsonld]
    elif isinstance(jsonld, str):
        return uuid_to_jsonld(jsonld)
    return generate_uuid()


def uuid_to_jsonld(uuid: str | None) -> str | None:
    """Removes the base IRI from an expanded @id."""
    if uuid is None:
        return None
    split = uuid.split(BASE_IRI)
    if len(split) == 1:
        return split[0]
    return BASE_IRI.join(split[1:])


def formatted_uuid_to_json(
    ctx: Context, uuid: None | str | list[str]
) -> str | dict[str, Any] | list[dict[str, Any]] | None:
    """Return a formatted node's uuid depending on the Croissant version."""
    if ctx.is_v0():
        return uuid  # type: ignore  # Force mypy types.
    else:
        if isinstance(uuid, list):
            if len(uuid) == 1:
                return {"@id": uuid[0]}
            else:
                return [{"@id": _uuid} for _uuid in uuid]
        elif isinstance(uuid, str):
            return {"@id": uuid}
        else:
            raise ValueError(f"Unknown uuid type: uuid is of type {type(uuid)}.")



"""Testing utils for `Operation`."""

from mlcroissant._src.operation_graph.base_operation import Operations


def operations() -> Operations:
    """An empty graph of operations to be used in tests."""
    return Operations()



from collections.abc import Iterable
import itertools
import re
from typing import Any


def regex_to_glob(regexes: str | list[str]) -> list[str]:
    """Converts a regular expression to several glob patterns.

    The function applies several transformations to achieve this:

    - Expand all non-capturing groups to possibly several expression. For example,
    for Hugging Face croissants, "default/(?:partial-)?train/.+parquet$" would become
    ["default/train/.+parquet$", "default/partial-train/.+parquet$"] to match both with
    and without the `partial-` prefix.

    - Convert to a glob pattern using some heuristics.
    """
    if isinstance(regexes, str):
        regexes = [regexes]
    for fn in [_expand_non_capturing_groups, _regex_to_glob_for_str]:
        regexes = list(itertools.chain.from_iterable(fn(regex) for regex in regexes))
    return regexes


def _expand_non_capturing_groups(regex: str) -> Iterable[str]:
    if "(?:" not in regex:
        # There is no non-capturing group:
        return [regex]

    # Find all capturing groups:
    pattern = r"\(\?:.*?\)\?|[^()]+"
    strings = re.findall(pattern, regex)
    if not strings:
        raise ValueError("the string should not be empty")
    subregex = "".join(strings[1:])
    # Recursively construct the results for the sub-regex:
    results = _expand_non_capturing_groups(subregex)
    string = strings[0]
    if string.startswith("(?:") and string.endswith(")?"):
        # Append the inside of the non-capturing group:
        string = string[len("(?:") : -len(")?")]
        return itertools.chain(
            [f"{string}{result}" for result in results],
            results,
        )
    else:
        # Append the string itself to each result:
        return [f"{string}{result}" for result in results]


def _regex_to_glob_for_str(regex: str) -> Iterable[str]:
    """Converts a regular expression to a glob pattern by unescaping regex syntax.

    Warning: this is based on manual heuristics to convert a regular expression to a
    glob expression.
    """
    # Remove starting ^
    regex = re.sub(r"^\^", "", regex)
    # Remove trailing $
    regex = re.sub(r"\$$", "", regex)
    # Interpret \. as .
    regex = re.sub(r"\\\.", ".", regex)
    # Interpret .* as *
    regex = re.sub(r"\.\*", "*", regex)
    # Interpret .+ as *
    regex = re.sub(r"\.\+", "*", regex)
    # Interpret \\- as -
    regex = re.sub(r"\\-", "-", regex)
    return [regex]


def capture_one_capturing_group(str_regex: str, value: Any) -> str:
    """Captures the one and only capturing group, but ignoring non-capturing gorups."""
    # Non-capturing groups have the form (?:a_non_capturing_group)
    capturing_groups = re.compile(r"\((?!\?\:).*?\)")
    groups = capturing_groups.findall(str_regex)
    if len(groups) == 1:
        matches = re.match(groups[0], value)
        if not matches:
            raise ValueError(
                "The replace value doesn't respect the expected capturing group."
                f" Expected: {groups[0]}. Got: {value}"
            )
    else:
        raise ValueError(
            "A transform regex should have exactly 1 capturing group in"
            f" the transform regex. Got: '{str_regex}'"
        )
    return capturing_groups.sub(re.escape(value), str_regex)



"""Tests for FileObjects."""

import pickle
from unittest import mock

from etils import epath
import pytest

from mlcroissant._src.core import constants
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.tests.nodes import create_test_node


@pytest.mark.parametrize(
    ["conforms_to", "field_uuid"],
    [[CroissantVersion.V_0_8, "name"], [CroissantVersion.V_1_0, "id"]],
)
def test_checks_are_performed(conforms_to, field_uuid):
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(
        Node, "assert_has_optional_properties"
    ), mock.patch.object(
        Node, "validate_name"
    ) as validate_name_mock, mock.patch.object(
        Node, "assert_has_exclusive_properties"
    ) as exclusive_mock:
        ctx = Context(conforms_to=conforms_to)
        create_test_node(FileObject, ctx=ctx)
        mandatory_mock.assert_has_calls([
            mock.call("encoding_format", field_uuid), mock.call("content_url")
        ])
        exclusive_mock.assert_called_once_with(["md5", "sha256"])
        validate_name_mock.assert_called_once()


@pytest.mark.parametrize(
    ["conforms_to", "field_uuid"],
    [[CroissantVersion.V_0_8, "name"], [CroissantVersion.V_1_0, "id"]],
)
def test_checks_not_performed_for_live_dataset(conforms_to, field_uuid):
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(
        Node, "assert_has_optional_properties"
    ), mock.patch.object(
        Node, "validate_name"
    ) as validate_name_mock, mock.patch.object(
        Node, "assert_has_exclusive_properties"
    ) as exclusive_mock:
        ctx = Context(is_live_dataset=True, conforms_to=conforms_to)
        create_test_node(FileObject, ctx=ctx)
        mandatory_mock.assert_has_calls([
            mock.call("encoding_format", field_uuid), mock.call("content_url")
        ])
        exclusive_mock.assert_not_called()
        validate_name_mock.assert_called_once()


@pytest.mark.parametrize(
    ["encoding"],
    [
        ["text/csv"],
        ["text/tab-separated-values"],
    ],
)
def test_from_jsonld(encoding):
    ctx = Context()
    jsonld = {
        "@type": constants.SCHEMA_ORG_FILE_OBJECT(ctx),
        "@id": "foo_id",
        constants.SCHEMA_ORG_NAME: "foo",
        constants.SCHEMA_ORG_DESCRIPTION: "bar",
        constants.SCHEMA_ORG_CONTENT_URL: "https://mlcommons.org",
        constants.SCHEMA_ORG_ENCODING_FORMAT: encoding,
        constants.SCHEMA_ORG_SHA256: (
            "48a7c257f3c90b2a3e529ddd2cca8f4f1bd8e49ed244ef53927649504ac55354"
        ),
    }
    ctx.mapping = {"file1": epath.Path("~/Downloads/file1.csv")}
    file_object = FileObject.from_jsonld(ctx, jsonld)
    assert file_object.name == "foo"
    assert file_object.id == "foo_id"
    assert file_object.description == "bar"
    assert file_object.content_url == "https://mlcommons.org"
    assert file_object.encoding_format == encoding
    assert (
        file_object.sha256
        == "48a7c257f3c90b2a3e529ddd2cca8f4f1bd8e49ed244ef53927649504ac55354"
    )
    assert not ctx.issues.errors


@pytest.mark.parametrize(
    ["conforms_to"],
    [[CroissantVersion.V_0_8], [CroissantVersion.V_1_0]],
)
def test_pickable(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    file_object = create_test_node(FileObject, ctx=ctx)
    file_object = pickle.loads(pickle.dumps(file_object))
    # Test that the context was successfully restored:
    assert file_object.ctx.conforms_to == conforms_to



"""Module to manage "bounding boxes" annotations on images."""

from typing import Any


def parse(value: Any) -> list[float]:
    """Parses a value to a machine-readable bounding box.

    Args:
        value: The value to parse can be either a single space-separated string or a
            list of float-compatible elements.

    Returns:
        The 4-float list that composes the bounding box.
    """
    if isinstance(value, list):
        pass
    elif isinstance(value, str):
        value = value.split(" ")
    else:
        raise ValueError(
            "Wrong format for a bounding box. Expected: str | list. Got:"
            f" {type(value)}. If you need to support more format, feel free to create"
            " an issue on GitHub."
        )
    try:
        value = [float(element) for element in value]
    except ValueError as e:
        raise ValueError(
            "Bounding boxes should have coordinates that can be converted to floats."
            f" Got {value}"
        ) from e
    if len(value) != 4:
        raise ValueError(
            "Bounding box could not be parsed. Bounding boxes should have a length of"
            f" 4. Got {len(value)}"
        )
    return value



"""RDF module to represent all things related to the Resource Description Framework."""

from __future__ import annotations

import dataclasses
import functools
from typing import Any

from absl import logging
from rdflib import term

from mlcroissant._src.core.types import Json


def get_context(json_: Json) -> Json:
    """Returns the context and raises an error if it is not a dictionary as expected."""
    context = json_.get("@context", {})
    if not isinstance(context, dict):
        raise ValueError("@context should be a dictionary. Got: {existing_context}")
    return context


def make_context(ctx=None, **kwargs):
    """Returns the JSON-LD @context with additional keys."""
    context = {
        "@language": "en",
        "@vocab": "https://schema.org/",
        "arrayShape": "cr:arrayShape" if ctx is not None and ctx.is_v1_1() else None,
        "citeAs": "cr:citeAs",
        "column": "cr:column",
        "conformsTo": "dct:conformsTo",
        "cr": "http://mlcommons.org/croissant/",
        "rai": "http://mlcommons.org/croissant/RAI/",
        "data": {"@id": "cr:data", "@type": "@json"},
        "dataType": {"@id": "cr:dataType", "@type": "@vocab"},
        "dct": "http://purl.org/dc/terms/",
        "examples": {"@id": "cr:examples", "@type": "@json"},
        "extract": "cr:extract",
        "field": "cr:field",
        "fileProperty": "cr:fileProperty",
        "fileObject": "cr:fileObject",
        "fileSet": "cr:fileSet",
        "format": "cr:format",
        "includes": "cr:includes",
        "isArray": "cr:isArray" if ctx is not None and ctx.is_v1_1() else None,
        "isLiveDataset": "cr:isLiveDataset",
        "jsonPath": "cr:jsonPath",
        "key": "sc:key" if ctx is not None and ctx.is_v0() else "cr:key",
        "md5": "sc:md5" if ctx is not None and ctx.is_v0() else "cr:md5",
        "parentField": "cr:parentField",
        "path": "cr:path",
        "recordSet": "cr:recordSet",
        "references": "cr:references",
        "regex": "cr:regex",
        "repeated": "cr:repeated",
        "replace": "cr:replace",
        "sc": "https://schema.org/",
        "separator": "cr:separator",
        "source": "cr:source",
        "subField": "cr:subField",
        "transform": "cr:transform",
        **kwargs,
    }
    return {key: value for key, value in context.items() if value is not None}


@dataclasses.dataclass(eq=False, repr=False)
class Rdf:
    """RDF-specific vocabulary on the metadata."""

    context: Json = dataclasses.field(default_factory=make_context)

    @classmethod
    def from_json(cls, ctx, json: Json) -> Rdf:
        """Creates a `Rdf` from JSON."""
        context = get_context(json)
        # Keys that are in the standard @context, but not in the current @context.
        different_keys = make_context(ctx).keys() - context.keys()
        if different_keys:
            logging.warning(
                "WARNING: The JSON-LD `@context` is not standard. Refer to the"
                " official @context (e.g., from the example datasets in"
                " https://github.com/mlcommons/croissant/tree/main/datasets/1.0). The"
                f" different keys are: {different_keys}"
            )
        return cls(context=context)

    @functools.cache
    def reverse_context(self) -> Json:
        """Reverses the context dictionary.

        - context = "ml"->"http://mlcommons.org/schema"
        - reverse_context = "http://mlcommons.org/schema"->"ml"
        """
        reversed_context = {}

        def add_mapping(k: str, v: Any):
            if not isinstance(v, str):
                return
            # cr:transform -> transform
            reversed_context[v] = k
            splits = v.split(":")
            abbreviation = splits[0]
            if url := self.abbreviations().get(abbreviation):
                if len(splits) > 2:
                    raise ValueError("Key in @context contains several colons (:)")
                value = f"{url}{splits[1]}"
                # https://mlcommons.org/croissant -> transform
                reversed_context[value] = k

        for k, v in self.context.items():
            if isinstance(v, str):
                add_mapping(k, v)
            elif isinstance(v, dict):
                add_mapping(k, v.get("@id"))
        return reversed_context

    @functools.cache
    def abbreviations(self) -> Json:
        """Lists all abbreviations, eg "ml"->"http://mlcommons.org/schema"."""
        return {
            k: v
            for k, v in reversed(self.context.items())
            if isinstance(v, str) and (v.startswith("https") or v.startswith("http"))
        }

    @functools.cache
    def shorten_value(self, value: str) -> str:
        """Shortens a value according to the context if possible."""
        for abbreviation, url in self.abbreviations().items():
            if value.startswith(url):
                return value.replace(url, f"{abbreviation}:")
        return value

    @functools.cache
    def shorten_key(self, key: str | term.URIRef) -> str:
        """Shortens a key according to the context if possible.

        Keys and values are shortened differently in Croissant. For example, as a value
        https://schema.org/FileObject will be `sc:FileObject`, but as a key
        https://schema.org/fileObject will be `fileObject`.
        """
        key = str(key)
        # 1. Try to shorten with the vocab:
        vocab = self.context.get("@vocab")
        if isinstance(vocab, str) and key.startswith(vocab):
            return key.replace(vocab, "", 1)
        # 2. Try to shorten with the context:
        if short := self.reverse_context().get(key):
            return short
        return self.shorten_value(key)



"""issues_test module."""

import textwrap

from mlcroissant._src.core.issues import Issues
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.metadata import Metadata


def test_issues():
    issues = Issues()
    assert not issues.errors
    assert not issues.warnings

    # With context
    metadata = Metadata(id="abc", name="abc")
    file_object = FileObject(id="xyz", name="xyz")
    file_object.parents = [metadata]
    issues.add_error("foo", metadata)
    issues.add_warning("bar", file_object)
    assert issues.errors == {"[Metadata(abc)] foo"}
    assert issues.warnings == {"[Metadata(abc) > FileObject(xyz)] bar"}

    # Without context
    issues.add_error("foo")
    issues.add_warning("bar")
    assert issues.errors == {
        "[Metadata(abc)] foo",
        "foo",
    }
    assert issues.warnings == {
        "[Metadata(abc) > FileObject(xyz)] bar",
        "bar",
    }

    # Final report
    assert issues.report() == textwrap.dedent(
        """Found the following 2 error(s) during the validation:
  -  [Metadata(abc)] foo
  -  foo
Found the following 2 warning(s) during the validation:
  -  [Metadata(abc) > FileObject(xyz)] bar
  -  bar"""
    )



"""Data operation module."""

import dataclasses

import pandas as pd

from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet


@dataclasses.dataclass(frozen=True, repr=False)
class Data(Operation):
    """Operation for inline data."""

    node: RecordSet

    def call(self, *args) -> pd.DataFrame:
        """See class' docstring."""
        del args  # unused
        return pd.DataFrame.from_records(self.node.data)



"""init_test module."""

from mlcroissant._src.operation_graph.operations import init
from mlcroissant._src.tests.nodes import empty_node
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = init.InitOperation(operations=operations(), node=empty_node)
    assert str(operation) == "InitOperation(node_name)"



"""Tests for dates."""

import datetime

import pytest

from mlcroissant._src.core.dates import from_datetime_to_str
from mlcroissant._src.core.dates import from_str_to_datetime
from mlcroissant._src.core.issues import Issues


@pytest.mark.parametrize(
    ["date", "expected", "errors"],
    [
        ["20240101", datetime.datetime(year=2024, month=1, day=1), set()],
        ["2024-01-01", datetime.datetime(year=2024, month=1, day=1), set()],
        [20240101, None, set(["Wrong type for date. Expected str. Got <class 'int'>"])],
        [
            "foo",
            None,
            set([
                "Dates or DateTimes should follow the [ISO 8601"
                " format](https://en.wikipedia.org/wiki/ISO_8601). Got foo"
            ]),
        ],
    ],
)
def test_from_str_to_datetime(date: str, expected: datetime.datetime, errors: set[str]):
    issues = Issues()
    result = from_str_to_datetime(issues, date)
    assert issues.errors == errors
    assert expected == result


@pytest.mark.parametrize(
    ["date", "expected"],
    [
        [None, None],
        [datetime.datetime(year=2024, month=1, day=1), "2024-01-01"],
        [
            datetime.datetime(year=2024, month=1, day=1, hour=12, minute=20),
            "2024-01-01T12:20:00",
        ],
    ],
)
def test_from_datetime_to_str(date: datetime.datetime | None, expected: str):
    assert from_datetime_to_str(date) == expected



"""Defines the public interface to the `mlcroissant.nodes` package."""

from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.metadata import Metadata
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.structure_graph.nodes.source import Transform

__all__ = [
    "Extract",
    "Field",
    "FileObject",
    "FileSet",
    "Metadata",
    "RecordSet",
    "Source",
    "Transform",
]



"""Field module."""

from rdflib import term
from rdflib.namespace import SDO
from typing_extensions import Self

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.data_types import data_types_from_jsonld
from mlcroissant._src.core.data_types import data_types_to_jsonld
from mlcroissant._src.core.data_types import EXPECTED_DATA_TYPES
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.base_node import node_by_uuid
from mlcroissant._src.structure_graph.nodes.source import Source


@mlc_dataclasses.dataclass
class ParentField(Node):
    """Class for the `parentField` property."""

    JSONLD_TYPE = None

    references: Source | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[Source],
        url=constants.ML_COMMONS_REFERENCES,
    )
    source: Source | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[Source],
        url=constants.ML_COMMONS_SOURCE,
    )


@mlc_dataclasses.dataclass
class Field(Node):
    """Nodes to describe a dataset Field."""

    JSONLD_TYPE = constants.ML_COMMONS_FIELD_TYPE

    array_shape: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "The shape of the array as a comma-separated string.  -1 indicates "
            " dimensions of unknown/unspecified size. (-1,) represents a simple list."
            " If specified,  then `is_array` must be True."
        ),
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_ARRAY_SHAPE,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_DESCRIPTION,
    )
    # `data_types` is different than `node.data_type`. See `data_type`'s docstring.
    data_types: list[term.URIRef] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "The data type of the field, identified by the URI of the corresponding"
            " class. It could be either an atomic type (e.g, `sc:Integer`) or a"
            " semantic type (e.g., `sc:GeoLocation`)."
        ),
        from_jsonld=data_types_from_jsonld,
        to_jsonld=data_types_to_jsonld,
        url=constants.ML_COMMONS_DATA_TYPE,
    )
    equivalentProperty: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "A property that is equivalent to this Field. Used in the case a dataType"
            " is specified on the RecordSet to map specific fields to specific"
            " properties associated with that dataType."
        ),
        input_types=[SDO.URL],
        url=constants.ML_COMMONS_EQUIVALENT_PROPERTY,
    )
    is_array: bool | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "If true, then the Field is an array of values of type dataType. If"
            " `array_shape` is not specified, it will default to (-1,), i.e. a"
            " one-dimensional array of unknown shape."
        ),
        input_types=[SDO.Boolean],
        url=constants.ML_COMMONS_IS_ARRAY,
    )
    is_enumeration: bool | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Boolean],
        url=constants.ML_COMMONS_IS_ENUMERATION,
        versions=[CroissantVersion.V_0_8],
    )
    name: str = mlc_dataclasses.jsonld_field(
        default="",
        description="The name of the Field.",
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_NAME,
    )
    parent_field: ParentField | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "A special case of `SubField` that should be hidden because it references a"
            " `Field` that already appears in the `RecordSet`."
        ),
        input_types=[ParentField],
        url=constants.ML_COMMONS_PARENT_FIELD,
    )
    references: Source = mlc_dataclasses.jsonld_field(
        default_factory=Source,
        description=(
            "Another `Field` of another `RecordSet` that this field references. This is"
            " the equivalent of a foreign key reference in a relational database."
        ),
        input_types=[Source],
        url=constants.ML_COMMONS_REFERENCES,
    )
    repeated: bool | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "[DEPRECATED]. Please use `is_array` and `array_shape`. If true, then the"
            " Field is a list of values of type dataType."
        ),
        input_types=[SDO.Boolean],
        url=constants.ML_COMMONS_REPEATED,
    )
    source: Source = mlc_dataclasses.jsonld_field(
        default_factory=Source,
        description=(
            "The data source of the field. This will generally reference a `FileObject`"
            " or `FileSet`'s contents (e.g., a specific column of a table)."
        ),
        input_types=[Source],
        url=constants.ML_COMMONS_SOURCE,
    )
    sub_fields: list[Self] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description="Another `Field` that is nested inside this one.",
        from_jsonld=lambda ctx, fields: Field.from_jsonld(ctx, fields),
        url=constants.ML_COMMONS_SUB_FIELD,
    )

    def __post_init__(self):
        """Checks arguments of the node."""
        Node.__post_init__(self)
        uuid_field = "name" if self.ctx.is_v0() else "id"
        self.validate_name()
        self.assert_has_mandatory_properties(uuid_field)
        self.source.check_source(self.add_error)
        self._standardize_data_types()
        if self.array_shape and not self.is_array:
            self.add_error(
                f"Field {self.uuid} defines `array_shape`, but `is_array` is False."
            )
        self.array_shape_tuple

    def _standardize_data_types(self):
        """Converts data_types to a list of rdflib.URIRef."""
        data_types = self.data_types
        if data_types is None:
            data_types = []
        if not isinstance(data_types, list):
            data_types = [data_types]
        self.data_types = [term.URIRef(data_type) for data_type in data_types]

    @property
    def array_shape_tuple(self) -> tuple[int, ...] | None:
        """Converts self.array_shape into a tuple of integers."""
        if self.is_array and not self.array_shape:
            return (-1,)
        elif self.array_shape:
            try:
                return tuple(int(dim) for dim in self.array_shape.split(","))
            except ValueError:
                self.add_error(
                    f"Field {self.uuid}'s attribute array_shape `{self.array_shape}`"
                    " cannot be parsed as a tuple of comma-separated values."
                )
        return None

    @property
    def data_type(self) -> type | term.URIRef | None:
        """Recursively retrieves the actual data type of the node.

        The data_type can be either directly on the node (`data_type`) or on one
        of the parent fields.

        `data_types` may contain semantic meaning, but `data_type` is the actual
        programmatic data type (i.e., bool, str, etc).
        """
        if self.sub_fields:
            return None
        data_types = self.data_types or []
        for data_type in data_types:
            # data_type can be matched to a Python type:
            if data_type in EXPECTED_DATA_TYPES:
                return EXPECTED_DATA_TYPES[term.URIRef(data_type)]
            # data_type is an ML semantic type:
            elif data_type in [
                DataType.IMAGE_OBJECT,
                # For some reasons, pytype cannot infer `Any` on ctx:
                DataType.BOUNDING_BOX,  # pytype: disable=wrong-arg-types
                DataType.AUDIO_OBJECT,
            ]:
                return term.URIRef(data_type)
        # The data_type has to be found on the source:
        source = node_by_uuid(self.ctx, self.source.uuid)
        if not isinstance(source, Field):
            self.add_error(
                "The field does not specify a valid"
                f" {constants.ML_COMMONS_DATA_TYPE(self.ctx)}, neither does any of"
                f" its predecessor. Got: {self.data_types}"
            )
            return None
        return source.data_type

    @property
    def data(self) -> str | None:
        """The data of the parent RecordSet."""
        if hasattr(self.parent, "data"):
            return getattr(self.parent, "data")
        return None

    @property
    def parent(self) -> Node:
        """Direct parent of the field."""
        parent = super().parent
        if parent:
            return parent
        raise ValueError(f"field={self} does not have any parent.")



"""Migration: http://mlcommons.org/schema/ -> http://mlcommons.org/schema/."""


def up(json_ld):
    """Up migration for http://mlcommons.org/schema/ -> http://mlcommons.org/schema/."""
    json_ld["@context"]["ml"] = "http://mlcommons.org/croissant/"
    return json_ld



"""Tests for Fields."""

from unittest import mock

import pytest
from rdflib import term

from mlcroissant._src.core.constants import DataType
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.tests.nodes import create_test_field
from mlcroissant._src.tests.nodes import create_test_node


@pytest.mark.parametrize(
    ["conforms_to", "field_uuid"],
    [[CroissantVersion.V_0_8, "name"], [CroissantVersion.V_1_0, "id"]],
)
def test_checks_are_performed(conforms_to, field_uuid):
    with mock.patch.object(
        Node, "assert_has_mandatory_properties"
    ) as mandatory_mock, mock.patch.object(Node, "validate_name") as validate_name_mock:
        ctx = Context(conforms_to=conforms_to)
        create_test_node(Field, ctx=ctx)
        mandatory_mock.assert_called_once_with(field_uuid)
        validate_name_mock.assert_called_once()


@pytest.mark.parametrize(
    ["array_shape", "array_shape_tuple"], [["1,2,3", (1, 2, 3)], [None, (-1,)]]
)
def test_array_shape_tuple(array_shape, array_shape_tuple):
    field = create_test_field(is_array=True, array_shape=array_shape)
    assert field.array_shape_tuple == array_shape_tuple


def test_data_type():
    # data_types can be a string:
    assert create_test_field(data_types=DataType.BOOL).data_types == [DataType.BOOL]
    # ...or a list of strings:
    assert create_test_field(
        data_types=[DataType.BOOL, "http://some-semantic-type"]
    ).data_types == [
        DataType.BOOL,
        term.URIRef("http://some-semantic-type"),
    ]

    # data_type are infered from the field...
    assert (
        create_test_field(
            data_types=[
                DataType.BOOL,
                "http://some-semantic-type",
            ]
        ).data_type
        is bool
    )
    # ...or from the predecessors. See the test case
    # `recordset_missing_context_for_datatype`.



"""Testing utils for `Node`."""

import functools
from typing import Callable, TypeVar

from etils import epath

from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import Issues
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet


class _EmptyNode(Node):
    def __post_init__(self):
        pass

    @classmethod
    def from_jsonld(cls):
        pass

    def to_json(self):
        pass


def _node_params(**kwargs):
    params = {
        "ctx": Context(folder=epath.Path()),
        "name": "node_name",
        "id": "node_name",
    }
    for key, value in kwargs.items():
        params[key] = value
    return params


T = TypeVar("T")


def create_test_node(cls: type[T], **kwargs) -> T:
    """Utils to easily create new nodes in tests.

    Usage:

    Instead of writing:
    ```python
    node = FileSet(
        id=...,
        issues=...,
        name=...,
        folder=...,
        description="Description"
    )
    ```

    Use:
    ```python
    node = create_test_file_set(description="Description")
    ```
    """
    return cls(**_node_params(**kwargs))


create_test_field: Callable[..., Field] = functools.partial(  # type: ignore  # Force mypy types.
    create_test_node, Field, name="field_name", id="field_name"
)
create_test_file_object: Callable[..., FileObject] = functools.partial(  # type: ignore  # Force mypy types.
    create_test_node, FileObject, name="file_object_name", id="file_object_name"
)
create_test_file_set: Callable[..., FileSet] = functools.partial(  # type: ignore  # Force mypy types.
    create_test_node, FileSet, name="file_set_name", id="file_set_name"
)
create_test_record_set: Callable[..., RecordSet] = functools.partial(  # type: ignore  # Force mypy types.
    create_test_node, RecordSet, name="record_set_name", id="record_set_name"
)


empty_field: Field = create_test_node(Field, name="field_name", id="field_name")
empty_file_object: FileObject = create_test_node(
    FileObject, name="file_object_name", id="file_object_name"
)
empty_file_set: FileSet = create_test_node(
    FileSet, name="file_set_name", id="file_set_name"
)
empty_node: Node = create_test_node(_EmptyNode)
empty_record_set: RecordSet = create_test_node(
    RecordSet, name="record_set_name", id="record_set_name"
)


def assert_contain_error(issues: Issues, error_msg: str):
    """Assert whether one of the errors in issues contains a message."""
    assert any(
        error_msg in error for error in issues.errors
    ), f'Could not find error "{error_msg}" in {list(issues.errors)}'


def assert_contain_warning(issues: Issues, error_msg: str):
    """Assert whether one of the warnings in issues contains a message."""
    assert any(
        error_msg in error for error in issues.warnings
    ), f'Could not find warning "{error_msg}" in {list(issues.warnings)}'



"""join_test module."""

from mlcroissant._src.operation_graph.operations import join
from mlcroissant._src.tests.nodes import empty_record_set
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = join.Join(operations=operations(), node=empty_record_set)
    assert str(operation) == "Join(record_set_name)"



"""base_node_test module."""

import dataclasses

import pytest
from rdflib.namespace import SDO

from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.structure_graph import base_node
from mlcroissant._src.tests.nodes import assert_contain_error
from mlcroissant._src.tests.nodes import assert_contain_warning
from mlcroissant._src.tests.nodes import create_test_node


@dataclasses.dataclass(eq=False, repr=False)
class Node(base_node.Node):
    property1: str = ""
    property2: str = ""

    @classmethod
    def from_jsonld(cls):
        pass

    def to_json(self):
        pass


def test_there_exists_at_least_one_property():
    node = create_test_node(
        Node,
        property1="property1",
        property2="property2",
    )
    assert node.there_exists_at_least_one_property(["property0", "property1"])
    assert not node.there_exists_at_least_one_property([])
    assert not node.there_exists_at_least_one_property(["property0"])


@pytest.mark.parametrize(
    ["name", "expected_errors", "conforms_to"],
    [
        [
            "a" * 256,
            [
                "The name"
                ' "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"'
                " is too long (>255 characters)."
            ],
            CroissantVersion.V_0_8,
        ],
        [
            "this is not valid",
            ['The name "this is not valid" contains forbidden characters.'],
            CroissantVersion.V_0_8,
        ],
        [
            {"not": {"a": {"string"}}},
            ["The name should be a string. Got: <class 'dict'>."],
            CroissantVersion.V_1_0,
        ],
    ],
)
def test_validate_name(name, expected_errors, conforms_to):
    node = create_test_node(Node, name=name, ctx=Context(conforms_to=conforms_to))
    node.validate_name()
    assert node.ctx.issues.errors
    for expected_error, error in zip(expected_errors, node.ctx.issues.errors):
        assert expected_error in error


@pytest.mark.parametrize(
    "conforms_to", [CroissantVersion.V_0_8, CroissantVersion.V_1_0]
)
def test_validate_correct_name(conforms_to):
    node = create_test_node(
        Node, name="a-regular-id", ctx=Context(conforms_to=conforms_to)
    )
    node.validate_name()
    assert not node.ctx.issues.errors


def test_validate_name_1_0():
    node = create_test_node(
        Node, name="this is not valid", ctx=Context(conforms_to=CroissantVersion.V_1_0)
    )
    node.validate_name()
    assert not node.ctx.issues.errors


def test_eq():
    node1 = create_test_node(Node, id="node1", name="node1")
    node2 = create_test_node(Node, id="node2", name="node2")
    node1_doppelganger = create_test_node(Node, id="node1", name="node1_doppelganger")
    # Same ID.
    assert node1 == node1_doppelganger
    # Different ID.
    assert node1 != node2


def test_custom_node_with_cardinality_one():
    @mlc_dataclasses.dataclass
    class CustomNode(base_node.Node):
        JSONLD_TYPE = "foo.org/CustomNode"

        property1: int | None = mlc_dataclasses.jsonld_field(
            cardinality="ONE",
            default=None,
            input_types=[SDO.Integer],
            url="foo.org/property1",
        )

    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": 42,
        },
    )
    assert node.property1 == 42
    assert not node.ctx.issues.errors
    assert node.to_json() == {
        "@id": "foo",
        "@type": "foo.org/CustomNode",
        "foo.org/property1": 42,
    }

    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": [42, 43],
        },
    )
    assert node.property1 == 42
    assert_contain_warning(
        node.ctx.issues, "`property1` has cardinality `ONE`, but got a list"
    )


def test_custom_node_with_cardinality_many():
    @mlc_dataclasses.dataclass
    class CustomNode(base_node.Node):
        JSONLD_TYPE = "foo.org/CustomNode"

        property1: list[int] = mlc_dataclasses.jsonld_field(
            cardinality="MANY",
            default_factory=list,
            input_types=[SDO.Integer],
            url="foo.org/property1",
        )

    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": [42, 43],
        },
    )
    assert node.property1 == [42, 43]
    assert not node.ctx.issues.errors
    assert node.to_json() == {
        "@id": "foo",
        "@type": "foo.org/CustomNode",
        "foo.org/property1": [42, 43],
    }

    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": 42,
        },
    )
    assert node.property1 == [42]
    assert node.to_json() == {
        "@id": "foo",
        "@type": "foo.org/CustomNode",
        "foo.org/property1": 42,
    }


def test_custom_node_with_input_types():
    @mlc_dataclasses.dataclass
    class ChildNode(base_node.Node):
        JSONLD_TYPE = "foo.org/ChildNode"

        child: str | None = mlc_dataclasses.jsonld_field(
            default=None,
            input_types=[SDO.Text],
            url="foo.org/child",
        )

    @mlc_dataclasses.dataclass
    class CustomNode(base_node.Node):
        JSONLD_TYPE = "foo.org/CustomNode"

        property1: int | None = mlc_dataclasses.jsonld_field(
            default=None,
            input_types=[SDO.Integer],
            url="foo.org/property1",
        )
        # pytype: disable=invalid-annotation, disable=name-error
        property2: list[ChildNode] = mlc_dataclasses.jsonld_field(
            cardinality="MANY",
            default_factory=list,
            input_types=[ChildNode],
            url="foo.org/property2",
        )
        # pytype: enable=invalid-annotation, enable=name-error

    # When from_jsonld succeeds:
    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": 42,
            "foo.org/property2": [
                {
                    "@type": "foo.org/ChildNode",
                    "@id": "child",
                    "foo.org/child": "this is the child",
                }
            ],
        },
    )
    assert node.property1 == 42
    assert len(node.property2) == 1
    assert node.property2[0].id == "child"
    assert node.property2[0].child == "this is the child"
    assert not node.ctx.issues.errors
    assert not node.ctx.issues.warnings
    assert node.to_json() == {
        "@type": "foo.org/CustomNode",
        "@id": "foo",
        "foo.org/property1": 42,
        "foo.org/property2": {
            "@type": "foo.org/ChildNode",
            "@id": "child",
            "foo.org/child": "this is the child",
        },
    }

    # When from_jsonld fails:
    node = CustomNode.from_jsonld(
        Context(),
        {
            "@type": "foo.org/CustomNode",
            "@id": "foo",
            "foo.org/property1": "this should be int",
            "foo.org/property2": ["this should be child node"],
        },
    )
    assert node.property1 == None
    assert node.property2 == []
    assert_contain_error(
        node.ctx.issues,
        "`property1` should have type https://schema.org/Integer, but got str",
    )
    assert_contain_error(
        node.ctx.issues, "`property2` should have type foo.org/ChildNode, but got str"
    )


def test_cast_fn():
    def node(cast_fn):
        @mlc_dataclasses.dataclass
        class Node(base_node.Node):
            JSONLD_TYPE = None

            field: int | None = mlc_dataclasses.jsonld_field(
                default=None,
                cast_fn=cast_fn,
                input_types=[SDO.Text],
                url="foo.org/property1",
            )

        return Node

    def cast_fn(value) -> int | None:
        del value
        return 42

    Node = node(cast_fn)
    assert Node(field="field").field == 42

    def cast_fn(value) -> int | None:
        del value
        raise ValueError("bad value")

    Node = node(cast_fn)
    assert Node(field="field").field == "field"
    assert_contain_error(Node(field="field").issues, "bad value")



"""Main entry point for mlcroissant CLI."""

import importlib
import sys

from absl import app


class Commands:
    """Possible commands."""

    DOCUMENTATION = "documentation"
    LOAD = "load"
    VALIDATE = "validate"


choices = set([
    Commands.DOCUMENTATION,
    Commands.LOAD,
    Commands.VALIDATE,
])


def main():
    """Main function for the CLI."""
    if len(sys.argv) < 2:
        raise ValueError(f"usage: mlcroissant {choices}")
    choice = sys.argv[1]
    if choice not in choices:
        raise ValueError(f"usage: mlcroissant {choices}. Got: `mlcroissant {choice}`")
    module = importlib.import_module(f"mlcroissant.scripts.{choice}")
    app.run(module.main)



"""Tests for execute."""

from unittest import mock

import pandas as pd
import pytest

from mlcroissant._src.core.issues import GenerationError
from mlcroissant._src.operation_graph.base_operation import Operations
from mlcroissant._src.operation_graph.execute import execute_downloads
from mlcroissant._src.operation_graph.execute import execute_operations_sequentially
from mlcroissant._src.operation_graph.operations import Data
from mlcroissant._src.operation_graph.operations import InitOperation
from mlcroissant._src.operation_graph.operations import ReadFields
from mlcroissant._src.operation_graph.operations.download import Download
from mlcroissant._src.tests.nodes import create_test_file_object
from mlcroissant._src.tests.nodes import create_test_record_set
from mlcroissant._src.tests.nodes import empty_record_set


def test_execute_downloads():
    operations = Operations()
    node1 = create_test_file_object(name="node1", id="node1")
    node2 = create_test_file_object(name="node2", id="node2")
    download1 = Download(operations=operations, node=node1)
    download2 = Download(operations=operations, node=node2)
    data = Data(operations=operations, node=empty_record_set)
    operations.add_node(download1)
    operations.add_node(download2)
    operations.add_node(download2)
    operations.add_node(data)
    with mock.patch.object(Download, "__call__") as download_call, mock.patch.object(
        Data, "__call__"
    ) as data_call:
        execute_downloads(operations)
        assert download_call.call_count == 2
        data_call.assert_not_called()


def test_only_execute_needed_operations():
    operations = Operations()
    node = create_test_file_object()
    record_set = create_test_record_set(id="my-record-set")
    init = InitOperation(operations=operations, node=node)
    (
        init
        >> Download(operations=operations, node=node)
        >> ReadFields(operations=operations, node=record_set)
    )

    # This operation is isolated in the operation graph and should not be executed:
    init >> Data(operations=operations, node=record_set)

    with mock.patch.object(Download, "__call__") as download_call, mock.patch.object(
        Data, "__call__"
    ) as data_call, mock.patch.object(
        ReadFields, "__call__", return_value=pd.Series([])
    ) as read_field_mock:
        # Use list(iterator) to actually yield all operations and execute them.
        list(execute_operations_sequentially("my-record-set", operations))
        assert download_call.call_count == 1
        assert read_field_mock.call_count == 1
        assert data_call.call_count == 0


def test_raises_with_an_explicit_mlcroissant_exception():
    with mock.patch.object(Download, "__call__", side_effect=ValueError):
        operations = Operations()
        node = create_test_file_object()
        record_set = create_test_record_set(id="my-record-set")
        (
            InitOperation(operations=operations, node=node)
            >> Download(operations=operations, node=node)
            >> ReadFields(operations=operations, node=record_set)
        )
        with pytest.raises(GenerationError, match=".+Download\\(file_object_name\\)"):
            list(execute_operations_sequentially("my-record-set", operations))



"""Metadata module."""

import datetime
from typing import Any

from etils import epath
from rdflib.namespace import SDO
import requests
from typing_extensions import Self

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.dates import cast_date
from mlcroissant._src.core.dates import cast_dates
from mlcroissant._src.core.issues import ValidationError
from mlcroissant._src.core.json_ld import expand_jsonld
from mlcroissant._src.core.json_ld import remove_empty_values
from mlcroissant._src.core.json_ld import sort_dict
from mlcroissant._src.core.rdf import Rdf
from mlcroissant._src.core.types import Json
from mlcroissant._src.core.url import is_url
from mlcroissant._src.core.versions import cast_version
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.graph import from_file_to_json
from mlcroissant._src.structure_graph.graph import from_nodes_to_graph
from mlcroissant._src.structure_graph.nodes.creative_work import CreativeWork
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.organization import Organization
from mlcroissant._src.structure_graph.nodes.person import Person
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet


@mlc_dataclasses.dataclass
class Metadata(Node):
    """Nodes to describe a dataset metadata."""

    JSONLD_TYPE = constants.SCHEMA_ORG_DATASET

    cite_as: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "A citation for a publication that describes the dataset. Ideally,"
            " citations should be expressed using the bibtex format. Note that this is"
            " different from schema.org/citation, which is used to make a citation to"
            " another publication from this dataset."
        ),
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_CITE_AS,
    )
    creators: list[Organization | Person] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description="The creator(s) of the dataset.",
        input_types=[Organization, Person],
        url=constants.SCHEMA_ORG_CREATOR,
    )
    date_created: datetime.datetime | None = mlc_dataclasses.jsonld_field(
        cast_fn=cast_date,
        default=None,
        description="The date the dataset was initially created.",
        input_types=[SDO.Date, SDO.DateTime],
        url=constants.SCHEMA_ORG_DATE_CREATED,
    )
    date_modified: datetime.datetime | None = mlc_dataclasses.jsonld_field(
        cast_fn=cast_date,
        default=None,
        description="The date the dataset was last modified.",
        input_types=[SDO.Date, SDO.DateTime],
        url=constants.SCHEMA_ORG_DATE_MODIFIED,
    )
    date_published: datetime.datetime | None = mlc_dataclasses.jsonld_field(
        cast_fn=cast_date,
        default=None,
        description="The date the dataset was published.",
        input_types=[SDO.Date, SDO.DateTime],
        url=constants.SCHEMA_ORG_DATE_PUBLISHED,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="Description of the dataset.",
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_DESCRIPTION,
    )
    keywords: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "A set of keywords associated with the dataset, either as free text, or"
            " a DefinedTerm with a formal definition."
        ),
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_KEYWORDS,
    )
    in_language: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description="The language(s) of the content of the dataset.",
        input_types=[SDO.Language, SDO.Text],
        url=SDO.inLanguage,
    )
    license: list[str | CreativeWork] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "The license of the dataset. Croissant recommends using the URL of a"
            " known license, e.g., one of the licenses listed at"
            " https://spdx.org/licenses/."
        ),
        input_types=[CreativeWork, SDO.Text, SDO.URL],
        url=constants.SCHEMA_ORG_LICENSE,
    )
    name: str = mlc_dataclasses.jsonld_field(
        default="",
        description="The name of the dataset.",
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_NAME,
    )
    publisher: list[Organization | Person] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "The publisher of the dataset, which may be distinct from its creator."
        ),
        input_types=[Organization, Person],
        url=constants.SCHEMA_ORG_PUBLISHER,
    )
    same_as: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "The URL of another Web resource that represents the same dataset as this"
            " one."
        ),
        input_types=[SDO.URL],
        url=constants.SCHEMA_ORG_SAME_AS,
    )
    sd_licence: list[str | CreativeWork] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "A license document that applies to this structured data, typically"
            " indicated by URL."
        ),
        input_types=[CreativeWork, SDO.Text, SDO.URL],
        url=SDO.sdLicense,
    )
    url: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "The URL of the dataset. This generally corresponds to the Web page for"
            " the dataset."
        ),
        input_types=[SDO.URL],
        url=constants.SCHEMA_ORG_URL,
    )
    version: str | None = mlc_dataclasses.jsonld_field(
        cast_fn=cast_version,
        default=None,
        description="The version of the dataset following the requirements below.",
        input_types=[SDO.Integer, SDO.Number, SDO.Text],
        url=constants.SCHEMA_ORG_VERSION,
    )
    distribution: list[FileObject | FileSet] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        input_types=[FileObject, FileSet],
        url=constants.SCHEMA_ORG_DISTRIBUTION,
    )
    record_sets: list[RecordSet] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        input_types=[RecordSet],
        url=constants.ML_COMMONS_RECORD_SET,
    )
    data_collection: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_COLLECTION,
    )
    data_collection_type: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_COLLECTION_TYPE,
    )
    data_collection_missing_data: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_COLLECTION_MISSING_DATA,
    )
    data_collection_raw_data: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_COLLECTION_RAW_DATA,
    )
    data_collection_timeframe: list[datetime.datetime] | None = (
        mlc_dataclasses.jsonld_field(
            cardinality="MANY",
            cast_fn=cast_dates,
            default=None,
            input_types=[SDO.Date, SDO.DateTime],
            url=constants.ML_COMMONS_RAI_DATA_COLLECTION_TIME_FRAME,
        )
    )
    data_imputation_protocol: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_IMPUTATION_PROTOCOL,
    )
    data_preprocessing_protocol: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_PREPROCESSING_PROTOCOL,
    )
    data_manipulation_protocol: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_DATA_MANIPULATION_PROTOCOL,
    )
    data_annotation_protocol: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_ANNOTATION_PROTOCOL,
    )
    data_annotation_platform: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_ANNOTATION_PLATFORM,
    )
    data_annotation_analysis: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_ANNOTATION_ANALYSIS,
    )
    annotations_per_item: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_ANNOTATIONS_PER_ITEM,
    )
    annotator_demographics: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_ANNOTATOR_DEMOGRAPHICS,
    )
    machine_annotation_tools: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_MACHINE_ANNOTATION_TOOLS,
    )
    data_biases: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_BIASES,
    )
    data_use_cases: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_USE_CASES,
    )
    data_limitations: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_LIMITATIONS,
    )
    data_social_impact: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_SOCIAL_IMPACT,
    )
    personal_sensitive_information: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_PERSONAL_SENSITIVE_INFORMATION,
    )
    data_release_maintenance_plan: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        input_types=[SDO.Text],
        url=constants.ML_COMMONS_RAI_DATA_RELEASE_MAINTENANCE_PLAN,
    )

    def _define_field_parents(self, fields: list[Field], parents: list[Any]):
        """Recursively populate the field's and subfield's parents."""
        for field in fields:
            field.parents = [self] + parents
            if field.sub_fields:
                field_parents = parents + [field]
                self._define_field_parents(
                    fields=field.sub_fields, parents=field_parents
                )

    def __post_init__(self):
        """Checks arguments of the node and setup ID."""
        Node.__post_init__(self)
        # Define parents.
        for node in self.distribution:
            node.parents = [self]
        for record_set in self.record_sets:
            record_set.parents = [self]
            self._define_field_parents(record_set.fields, parents=[record_set])

        # Back-fill the graph in every node.
        self.ctx.graph = from_nodes_to_graph(self)
        self.check_graph()

        # Check properties.
        self.validate_name()
        self.license = self.validate_license()
        if self.ctx.is_v0():
            self.assert_has_mandatory_properties("name")
        self.assert_has_optional_properties(
            "cite_as", "date_published", "license", "version"
        )

        # Raise exception if there are errors.
        for node in self.nodes():
            if node.ctx.issues.errors:
                raise ValidationError(node.ctx.issues.report())

        self.ctx.conforms_to = CroissantVersion.from_jsonld(
            self.ctx, self.ctx.conforms_to
        )

        # Share the structure graph in the context
        for node in self.nodes():
            node.ctx.graph = self.ctx.graph

    def to_json(self) -> Json:
        """Converts the `Metadata` to JSON."""
        context = self.ctx.rdf.context
        if "@base" in context:
            if context["@base"] == constants.BASE_IRI:
                context.pop("@base")
        jsonld = super().to_json()
        jsonld.pop("@id", None)
        jsonld["@context"] = context
        conforms_to = self.ctx.conforms_to.to_json() if self.ctx.conforms_to else None
        jsonld["conformsTo"] = conforms_to
        if self.ctx.is_live_dataset:
            jsonld["isLiveDataset"] = self.ctx.is_live_dataset
        jsonld = remove_empty_values(jsonld)
        return sort_dict(jsonld)

    @property
    def file_objects(self) -> list[FileObject]:
        """Gets `https://schema.org/FileObject` in the distribution."""
        return [
            file_object
            for file_object in self.distribution
            if isinstance(file_object, FileObject)
        ]

    @property
    def file_sets(self) -> list[FileSet]:
        """Gets `https://schema.org/FileSet` in the distribution."""
        return [
            file_set for file_set in self.distribution if isinstance(file_set, FileSet)
        ]

    def nodes(self) -> list[Node]:
        """List all nodes in metadata."""
        nodes: list[Node] = [self]
        nodes.extend(self.distribution)
        nodes.extend(self.record_sets)
        for record_set in self.record_sets:
            nodes.extend(record_set.fields)
            for field in record_set.fields:
                nodes.extend(field.sub_fields)
        return nodes

    def validate_license(self) -> list[str | CreativeWork] | None:
        """Validates the license as a list of strings."""
        license = self.license
        input_types = (str, CreativeWork)
        if license is None:
            return None
        elif isinstance(license, list):
            if any(not isinstance(el, input_types) for el in license):
                self.add_error(
                    f"License should be a list of str or CreativeWork. Got: {license}"
                )
                return None
            return license
        elif isinstance(license, input_types):
            return [license]
        else:
            self.add_error(
                f"License should be a list of str or CreativeWork. Got: {license}"
            )
            return None

    def check_graph(self):
        """Checks the integrity of the structure graph.

        The rules are the following:
        - The graph is directed.
        - All fields have a data type: either directly specified, or on a parent.

        Args:
            issues: The issues to populate in case of problem.
            graph: The structure graph to be checked.
        """
        # Check that the graph is directed.
        if not self.ctx.graph.is_directed():
            self.ctx.issues.add_error("The structure graph is not directed.")
        fields = [node for node in self.ctx.graph.nodes if isinstance(node, Field)]
        # Check all fields have a data type: either on the field, on a parent.
        for field in fields:
            field.data_type

    @classmethod
    def from_file(cls, ctx: Context, file: epath.PathLike) -> Self:
        """Creates the Metadata from a Croissant file."""
        if is_url(file):
            response = requests.get(file)
            json_ = response.json()
            return cls.from_json(ctx=ctx, json_=json_)
        folder, json_ = from_file_to_json(file)
        ctx.folder = folder
        return cls.from_json(ctx=ctx, json_=json_)

    @classmethod
    def from_json(
        cls,
        ctx: Context,
        json_: Json,
    ) -> Self:
        """Creates a `Metadata` from JSON."""
        ctx.rdf = Rdf.from_json(ctx, json_)
        jsonld = expand_jsonld(json_, ctx=ctx)
        return cls.from_jsonld(ctx=ctx, jsonld=jsonld)



"""Base operation module."""

from __future__ import annotations

import abc
import dataclasses
import types
from typing import Any, Generic, Iterable, Sequence, TypeVar

import networkx as nx
import pandas as pd

from mlcroissant._src.structure_graph.base_node import Node


class Operations(nx.DiGraph):
    """Overwrites nx.DiGraph to keep track of operations.

    `self.last_operations` maintains a pointer to the chain of last operations for each
    node.
    """

    def __init__(self):
        """Initializes `self.last_operations`."""
        super().__init__(self)
        self.last_operations: dict[Node, Operation] = {}

    def add_node(self, operation: Operation) -> None:
        """Overloads nx.add_node to keep track of the last operations."""
        if not self.has_node(operation):
            super().add_node(operation)

    def add_edge(self, operation1: Operation, operation2: Operation) -> None:
        """Overloads nx.add_node to keep track of the last operations."""
        if not self.has_edge(operation1, operation2):
            super().add_edge(operation1, operation2)

    # pytype: disable=signature-mismatch
    def predecessors(self, operation: Operation) -> Iterable[Operation]:
        """Overloads nx.predecessors to have types."""
        return super().predecessors(operation)

    # pytype: enable=signature-mismatch

    @property
    def nodes(self) -> Iterable[Operation]:
        """Overloads nx.nodes to return an interator of operations."""
        return super().nodes()

    def entry_operations(self) -> Sequence[Operation]:
        """Lists all operations without a parent in the graph of operations."""
        return [
            operation
            for operation, indegree in self.in_degree(self.nodes)
            if indegree == 0 and isinstance(operation, Operation)
        ]


OutputT = TypeVar("OutputT")


@dataclasses.dataclass(frozen=True, repr=False)
class Operation(abc.ABC, Generic[OutputT]):
    """Generic base class to define an operation.

    `@dataclass(frozen=True)` allows having a hashable operation for NetworkX to use
    operations as nodes of  graphs.

    `@dataclass(repr=False)` allows having a readable stringified `str(operation)`.

    `_output` is an internal-only field used to cache the output of calling the
    operation (`self.call()`).

    Args:
        operations: The graph of all operations.
        node: The node attached to the operation for the context.
    """

    operations: Operations
    node: Node
    _output: OutputT = dataclasses.field(init=False, hash=False, compare=False)

    def __post_init__(self):
        """Adds the operation to the graph of operations."""
        operation = self
        # If the operation is already in the graph we don't want to leak other
        # references to the same operation, so we find the original operation:
        if self in self.operations:
            operation = next(
                operation for operation in self.operations if operation == self
            )
        operation.operations.add_node(operation)
        operation.connect_to_last_operation(operation.node)
        operation.set_last_operation_for(operation.node)

    def set_last_operation_for(self, node: Node):
        """Sets self as the last operation for the node and its successors."""
        self.operations.last_operations[node] = self
        for successor in node.successors:
            # Report the last operation to the next node in the graph.
            self.operations.last_operations[successor] = self

    def connect_to_last_operation(self, node: Node):
        """Connects the current operation (self) to the last operation for node."""
        # The previous operation can be indexed either on the node itself...
        if node in self.operations.last_operations:
            previous_operation = self.operations.last_operations[node]
            if previous_operation != self:
                self.operations.add_edge(previous_operation, self)
        # ...or the node's predecessor.
        elif node.predecessor in self.operations.last_operations:
            previous_operation = self.operations.last_operations[node.predecessor]
            if previous_operation != self:
                self.operations.add_edge(previous_operation, self)

    def __call__(self, set_output_in_memory: bool = False) -> OutputT:
        """Executes the current operation from the output of its parents and store the result."""
        if self.has_output():
            return self._output
        inputs = self.inputs
        output = self.call() if inputs is None else self.call(*inputs)
        if isinstance(output, types.GeneratorType) and set_output_in_memory:
            output = pd.DataFrame(output)  # type:ignore
        self.set_output(output)
        return output

    @abc.abstractmethod
    def call(self, *args, **kwargs) -> OutputT:
        """Abstract method to implement when an operation is called."""
        raise NotImplementedError

    @property
    def inputs(self) -> Sequence[Any]:
        """Returns the inputs of the operation - i.e. the outputs of the predecessors."""
        inputs = []
        parents = self.operations.predecessors(self)
        for parent in parents:
            if not parent.has_output():
                raise ValueError(
                    f'did not set output for "{parent}". This situation should not'
                    " happen as we go through the graph in a topological sort."
                )
            inputs.append(parent._output)
        return inputs

    @property
    def output(self) -> OutputT:
        """Returns the cached output of the operation if it is set."""
        return self._output

    def has_output(self) -> bool:
        """Checks whether self.output was already set."""
        return hasattr(self, "_output")

    def set_output(self, output: OutputT):
        """Sets the given output."""
        object.__setattr__(self, "_output", output)

    def __repr__(self):
        """Prints a simplified string representation of the operation."""
        return f"{type(self).__name__}({self.node.uuid})"

    def __rrshift__(
        self, left_operations: Operation | list[Operation] | None
    ) -> Operation:
        """Allows to chain operations in their graph: `operation1 >> operation2`."""
        right_operation = self
        if isinstance(left_operations, Operation):
            left_operations = [left_operations]
        elif left_operations is None:
            left_operations = []
        for left_operation in left_operations:
            self.operations.add_edge(left_operation, right_operation)
        return right_operation

    def __reduce__(self):
        """Allows pickling the operation.

        `self.operations` is stored separately in the state to break the cyclic
        reference between Operation and Operations. We could refactor the codebase to
        not have this dependency, but it would be a bigger change and it's convenient
        to be able to reference all operations from a single operation.
        """
        state = self.__getstate__()
        state_setter = {"operations": self.operations}
        if "_output" in state:
            state_setter["_output"] = state.pop("_output")
        args = tuple(state.values())
        return (
            self.__class__,
            args,
            state_setter,
        )

    def __getstate__(self):
        """Overwrites __getstate__ for pickling."""
        state = {}
        for field in dataclasses.fields(self):
            if field.name == "operations":
                state[field.name] = Operations()
            elif hasattr(self, field.name):  # _output may not be set
                state[field.name] = getattr(self, field.name)
        return state

    def __setstate__(self, state):
        """Overwrites __setstate__ for pickling."""
        object.__setattr__(self, "operations", state["operations"])
        if "_output" in state:
            self.set_output(state["_output"])



"""Test utils to handle records."""

import math
import re
from typing import Any

import pandas as pd


def record_to_python(record: Any):
    """Converts a record to a fully Python-native object.

    Warning: this function must be used for testing-purposes only!

    Records may contain non-serializable values (like `nan` or `pd.Timestamp` for
    example). This util converts records to Python-native objects:
    - bytes -> str
    - nan -> None
    - pd.Timestamp -> pd.Timestamp.strftime
    """
    if isinstance(record, bytes):
        return record.decode()
    elif isinstance(record, pd.Timestamp):
        return record.strftime("%Y-%m-%d %X")
    elif isinstance(record, float) and math.isnan(record):
        return None
    elif isinstance(record, (bool, float, int)):
        return record
    elif not isinstance(record, dict):
        str_repr = str(record)
        # Remove memory addresses from string representation.
        memory_addresses = re.compile(r"0x[0-9a-fA-F]+")
        return memory_addresses.sub("<MEMORY_ADDRESS>", str_repr)
    else:
        # Record is a dict
        for key, value in record.items():
            record[key] = record_to_python(value)
        return record



"""graph module."""

import dataclasses

from etils import epath
import networkx as nx

from mlcroissant._src.core.constants import EncodingFormat
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.issues import Issues
from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.base_operation import Operations
from mlcroissant._src.operation_graph.operations import Concatenate
from mlcroissant._src.operation_graph.operations import Data
from mlcroissant._src.operation_graph.operations import Download
from mlcroissant._src.operation_graph.operations import Extract
from mlcroissant._src.operation_graph.operations import FilterFiles
from mlcroissant._src.operation_graph.operations import InitOperation
from mlcroissant._src.operation_graph.operations import Join
from mlcroissant._src.operation_graph.operations import LocalDirectory
from mlcroissant._src.operation_graph.operations import Read
from mlcroissant._src.operation_graph.operations import ReadFields
from mlcroissant._src.operation_graph.operations.extract import should_extract
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.base_node import node_by_uuid
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet


def _find_record_set(node: Node) -> RecordSet:
    """Finds the record set to which a field is attached.

    The record set will be typically either the parent or the parent's parent.
    """
    for parent in reversed(node.parents):
        if isinstance(parent, RecordSet):
            return parent
    raise ValueError(f"Node {node} has no RecordSet parent.")


def _add_operations_for_record_set(
    operations: Operations,
    record_set: RecordSet,
):
    """Adds all operations for a node of type `RecordSet`."""
    if record_set.data:
        Data(operations=operations, node=record_set) >> ReadFields(
            operations=operations, node=record_set
        )
    has_join = any(field for field in record_set.fields if field.references.uuid)
    if has_join:
        Join(operations=operations, node=record_set) >> ReadFields(
            operations=operations, node=record_set
        )
    else:
        ReadFields(operations=operations, node=record_set)


def _add_operations_for_file_object(
    operations: Operations,
    node: FileObject,
    folder: epath.Path,
):
    """Adds all operations for a node of type `FileObject`.

    Operations are:

    - `Download`.
    - `Extract` if the file needs to be extracted.
    - `Concatenate` to merge several dataframes into one.
    - `Read` to read the file if it's a CSV.
    """
    operation: Operation | None = None
    if not node.contained_in:
        # Download the file
        operation = Download(operations=operations, node=node)
    first_operation = operation
    for successor in node.successors:
        # Reset `operation` to be the very first operation at each loop.
        operation = first_operation
        # Extract the file if needed
        if (
            should_extract(node.encoding_format)
            and isinstance(successor, (FileObject, FileSet))
            and not should_extract(successor.encoding_format)
        ):
            operation = operation >> Extract(operations=operations, node=node)
        if isinstance(successor, FileSet):
            operation = (
                operation
                >> FilterFiles(operations=operations, node=successor)
                >> Concatenate(operations=operations, node=successor)
            )
        if node.encoding_format and not should_extract(node.encoding_format):
            fields = tuple([
                field for field in node.recursive_successors if isinstance(field, Field)
            ])
            operation >> Read(
                operations=operations,
                node=node,
                folder=folder,
                fields=fields,
            )


def _add_operations_for_git(
    operations: Operations,
    node: FileObject,
    folder: epath.Path,
):
    """Adds all operations for a FileObject reading from a Git repository."""
    operation = Download(operations=operations, node=node)
    for successor in node.successors:
        if isinstance(successor, FileSet):
            fields = tuple(
                field
                for field in successor.recursive_successors
                if isinstance(field, Field)
            )
            (
                operation
                >> FilterFiles(operations=operations, node=successor)
                >> Read(
                    operations=operations,
                    node=successor,
                    folder=folder,
                    fields=fields,
                )
            )


def _add_operations_for_local_file_sets(
    operations: Operations,
    node: FileSet,
    folder: epath.Path,
):
    """Adds all operations for a FileSet reading from local files."""
    fields = tuple([
        field for field in node.recursive_successors if isinstance(field, Field)
    ])
    (
        LocalDirectory(
            operations=operations,
            node=node,
            folder=folder,
        )
        >> FilterFiles(operations=operations, node=node)
        >> Read(
            operations=operations,
            node=node,
            folder=folder,
            fields=fields,
        )
    )


def _add_operations_for_field(operations: Operations, node: Field):
    """Adds all joins for a Field."""
    record_set = _find_record_set(node)
    joins = [
        operation
        for operation in operations.nodes
        if isinstance(operation, Join) and operation.node == record_set
    ]
    for join in joins:
        left_node = node_by_uuid(node.ctx, node.source.uuid)
        right_node = node_by_uuid(node.ctx, node.references.uuid)
        if left_node and right_node:
            join.connect_to_last_operation(left_node)
            join.connect_to_last_operation(right_node)


@dataclasses.dataclass(frozen=True)
class OperationGraph:
    """Graph of dependent operations to execute to generate the dataset."""

    issues: Issues
    operations: Operations

    @classmethod
    def from_nodes(cls, ctx: Context, metadata: Node) -> "OperationGraph":
        """Builds the ComputationGraph from the nodes.

        This is done by:

        1. Building the structure graph.
        2. Building the computation graph by exploring the structure graph layers by
        layers in a breadth-first search.
        """
        operations = Operations()
        for node in nx.topological_sort(ctx.graph):
            if isinstance(node, FileObject):
                if node.encoding_format == EncodingFormat.GIT:
                    _add_operations_for_git(operations, node, ctx.folder)
                else:
                    _add_operations_for_file_object(operations, node, ctx.folder)
            elif isinstance(node, FileSet):
                if not node.contained_in:
                    _add_operations_for_local_file_sets(operations, node, ctx.folder)
            elif isinstance(node, RecordSet):
                _add_operations_for_record_set(
                    operations,
                    node,
                )
            elif isinstance(node, Field):
                _add_operations_for_field(operations, node)

        # Attach all entry nodes to a single `start` node
        entry_operations = operations.entry_operations()
        init_operation = InitOperation(operations=operations, node=metadata)
        for entry_operation in entry_operations:
            operations.add_edge(init_operation, entry_operation)
        return OperationGraph(issues=ctx.issues, operations=operations)

    def check_graph(self, ctx: Context):
        """Checks the operation graph for issues."""
        if not self.operations.is_directed():
            self.issues.add_error("Computation graph is not directed.")
        selfloops = [operation for operation, _ in nx.selfloop_edges(self.operations)]
        if selfloops:
            self.issues.add_error(
                f"The following operations refered to themselves: {selfloops}"
            )
        record_sets = [node for node in ctx.graph.nodes if isinstance(node, RecordSet)]
        for record_set in record_sets:
            record_set.check_joins_in_fields()



"""local_directory_test module."""

from etils import epath

from mlcroissant._src.operation_graph.operations.local_directory import LocalDirectory
from mlcroissant._src.tests.nodes import empty_file_set
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = LocalDirectory(
        operations=operations(), node=empty_file_set, folder=epath.Path("/foo/bar")
    )
    assert str(operation) == "LocalDirectory(file_set_name)"



"""Migration: https://github.com/mlcommons/croissant/issues/200."""


def _migrate_field(json_ld):
    for field in ["source", "references"]:
        if field in json_ld:
            source = json_ld[field]
            if "extract" in source:
                extract = source["extract"]
                if "csvColumn" in extract:
                    extract["column"] = extract["csvColumn"]
                    del extract["csvColumn"]
                    source["extract"] = extract
            json_ld[field] = source
    return json_ld


def up(json_ld):
    """Up migration that migrates `csvColumn`->`column`."""
    json_copy = json_ld.copy()
    for i, record_set in enumerate(json_copy.get("recordSet", [])):
        fields = record_set.get("field", [])
        if isinstance(fields, dict):
            fields = [fields]
        for j, field in enumerate(fields):
            field = _migrate_field(field)
            sub_fields = field.get("subField", [])
            if isinstance(sub_fields, dict):
                sub_fields = [sub_fields]
            for k, sub_field in enumerate(sub_fields):
                sub_field = _migrate_field(sub_field)
                field["subField"][k] = sub_field
            record_set["field"][j] = field
        json_ld["recordSet"][i] = record_set
    json_ld["@context"]["column"] = "ml:column"
    if "csvColumn" in json_ld["@context"]:
        del json_ld["@context"]["csvColumn"]
    return json_ld



"""Person module."""

from rdflib.namespace import SDO

from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.structure_graph.base_node import Node


@mlc_dataclasses.dataclass
class Person(Node):
    """Represents schema.org/Person."""

    JSONLD_TYPE = SDO.Person

    name: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="The name of the item.",
        input_types=[SDO.Text],
        url=SDO.name,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="A description of the item.",
        input_types=[SDO.Text],
        url=SDO.description,
    )
    email: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="Email address.",
        input_types=[SDO.Text],
        url=SDO.email,
    )
    url: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="URL of the item.",
        input_types=[SDO.URL],
        url=SDO.url,
    )



"""migrate_test module."""

import json

from etils import epath
import pytest

from mlcroissant._src.core.rdf import make_context
from mlcroissant._src.datasets import Dataset


# If this test fails, you probably manually updated a dataset in datasets/.
# Please, use scripts/migrations/migrate.py to migrate datasets.
@pytest.mark.parametrize("version", ["0.8", "1.0", "1.1"])
def test_expand_and_reduce_json_ld(version):
    dataset_folder = (
        epath.Path(__file__).parent.parent.parent.parent.parent.parent
        / "datasets"
        / version
    )
    paths = [path for path in dataset_folder.glob("*/*.json")]
    assert paths, f"Warning: Checking an empty list of paths: {dataset_folder}"
    for path in paths:
        print(f"Test for {path}")
        with path.open() as f:
            expected = json.load(f)
        dataset = Dataset(path)
        metadata = dataset.metadata
        actual = metadata.to_json()
        assert actual == expected, f"Error in {path}"


def test_make_context():
    assert make_context(foo="bar") == {
        "@language": "en",
        "@vocab": "https://schema.org/",
        "citeAs": "cr:citeAs",
        "column": "cr:column",
        "conformsTo": "dct:conformsTo",
        "cr": "http://mlcommons.org/croissant/",
        "rai": "http://mlcommons.org/croissant/RAI/",
        "data": {"@id": "cr:data", "@type": "@json"},
        "dataType": {"@id": "cr:dataType", "@type": "@vocab"},
        "dct": "http://purl.org/dc/terms/",
        "examples": {"@id": "cr:examples", "@type": "@json"},
        "extract": "cr:extract",
        "field": "cr:field",
        "fileProperty": "cr:fileProperty",
        "fileObject": "cr:fileObject",
        "fileSet": "cr:fileSet",
        "format": "cr:format",
        "includes": "cr:includes",
        "isLiveDataset": "cr:isLiveDataset",
        "jsonPath": "cr:jsonPath",
        "key": "cr:key",
        "md5": "cr:md5",
        "parentField": "cr:parentField",
        "path": "cr:path",
        "recordSet": "cr:recordSet",
        "references": "cr:references",
        "regex": "cr:regex",
        "repeated": "cr:repeated",
        "replace": "cr:replace",
        "sc": "https://schema.org/",
        "separator": "cr:separator",
        "source": "cr:source",
        "subField": "cr:subField",
        "transform": "cr:transform",
        "foo": "bar",
    }



"""regex_test module."""

import pytest

from mlcroissant._src.core import regex as regex_lib


@pytest.mark.parametrize(
    ["regex", "output"],
    [
        [
            "(?:baz)?xxx(?:foo)?yyy(?:bar)?",
            [
                "bazxxxfooyyybar",
                "bazxxxfooyyy",
                "bazxxxyyybar",
                "bazxxxyyy",
                "xxxfooyyybar",
                "xxxfooyyy",
                "xxxyyybar",
                "xxxyyy",
            ],
        ],
        [
            "^.+/train/.*\.parquet$",  # From a valid regex...
            [
                "*/train/*.parquet",  # ...to a valid glob pattern.
            ],
        ],
        [
            "^.+/my\\-train/.*\.parquet$",  # From a valid regex...
            [
                "*/my-train/*.parquet",  # ...to a valid glob pattern.
            ],
        ],
    ],
)
def test_regex_to_glob(regex: str, output: list[str]):
    assert regex_lib.regex_to_glob(regex) == output


def test_capture_one_capturing_group():
    # The value does not match:
    with pytest.raises(ValueError):
        regex_lib.capture_one_capturing_group(
            "default/(?:partial-)?(train|test)/.+parquet$", "NOT MATCHING"
        )

    # Too many capturing groups:
    with pytest.raises(ValueError):
        regex_lib.capture_one_capturing_group(
            "(default)/(?:partial-)?(train|test)/.+parquet$", "NOT MATCHING"
        )

    assert (
        regex_lib.capture_one_capturing_group(
            "default/(?:partial-)?(train|test)/.+parquet$", "train"
        )
        == "default/(?:partial-)?train/.+parquet$"
    )



"""parse_json_test module."""

import pandas as pd

from mlcroissant._src.operation_graph.operations.parse_json import parse_json_content
from mlcroissant._src.structure_graph.nodes.source import Extract
from mlcroissant._src.structure_graph.nodes.source import Source
from mlcroissant._src.tests.nodes import create_test_field


def test_parse_json():
    field1 = create_test_field(
        source=Source(extract=Extract(json_path="$.annotations[*].id"))
    )
    field2 = create_test_field(
        source=Source(extract=Extract(json_path="$.annotations[*].value"))
    )
    fields = (field1, field2)
    json = {
        "foo": "bar",
        "annotations": [
            {"id": 1, "value": 3},
            {"id": 2, "value": 4},
        ],
    }
    expected_df = pd.DataFrame(
        data={"$.annotations[*].id": [1, 2], "$.annotations[*].value": [3, 4]}
    )
    pd.testing.assert_frame_equal(parse_json_content(json, fields), expected_df)



"""FileSet module."""

from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.uuid import formatted_uuid_to_json
from mlcroissant._src.core.uuid import uuid_from_jsonld
from mlcroissant._src.structure_graph.base_node import Node


@mlc_dataclasses.dataclass
class FileSet(Node):
    """Nodes to describe a dataset FileSet (distribution)."""

    JSONLD_TYPE = constants.SCHEMA_ORG_FILE_SET

    contained_in: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "Another FileObject or FileSet that this one is contained in, e.g., in the"
            " case of a file extracted from an archive. When this property is present,"
            " the contentUrl is evaluated as a relative path within the container"
            " object"
        ),
        from_jsonld=lambda ctx, contained_in: uuid_from_jsonld(contained_in),
        to_jsonld=lambda ctx, contained_in: [
            formatted_uuid_to_json(ctx, uuid) for uuid in contained_in
        ],
        url=SDO.containedIn,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=SDO.description,
    )
    encoding_format: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="The format of the file, given as a mime type.",
        input_types=[SDO.Text],
        url=SDO.encodingFormat,
    )
    # TODO(https://github.com/mlcommons/croissant/issues/772): Support in mlcroissant.
    excludes: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "A glob pattern that specifies the files to exclude. Warning: This feature"
            " is not implemented yet in mlcroissant, so this is a no-op."
        ),
        input_types=[SDO.Text],
        url=lambda ctx: constants.ML_COMMONS_EXCLUDES(ctx),
    )
    includes: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description="A glob pattern that specifies the files to include.",
        input_types=[SDO.Text],
        url=lambda ctx: constants.ML_COMMONS_INCLUDES(ctx),
    )
    name: str = mlc_dataclasses.jsonld_field(
        default="",
        description=(
            "The name of the file.  As much as possible, the name should reflect the"
            " name of the file as downloaded, including the file extension. e.g."
            ' "images.zip".'
        ),
        input_types=[SDO.Text],
        url=SDO.name,
    )

    def __post_init__(self):
        """Checks arguments of the node."""
        Node.__post_init__(self)
        uuid_field = "name" if self.ctx.is_v0() else "id"
        self.validate_name()
        self.assert_has_mandatory_properties("includes", "encoding_format", uuid_field)



"""Structure graph module.

The goal of this module is the static analysis of the JSON file. We convert the initial
JSON to a so-called "structure graph", which is a Python representation of the JSON
containing the nodes (Metadata, FileObject, etc) and the hierarchy between them. In the
process of parsing all the nodes, we also check that no information is missing and raise
issues (errors or warnings) when necessary. See the docstring of
`from_nodes_to_structure_graph` for more information.

The important functions of this module are:
- from_file_to_jsonld           file -> JSON-LD
- from_nodes_to_graph           Metadata -> graph

TODO(https://github.com/mlcommons/croissant/issues/114):
A lot of methods in this file share common data structures (issues, graph, folder, etc),
so they should be grouped under a common `StructureGraph` class.
"""

from __future__ import annotations

import json

from etils import epath
import networkx as nx

from mlcroissant._src.core import constants
from mlcroissant._src.core.types import Json
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.file_object import FileObject
from mlcroissant._src.structure_graph.nodes.file_set import FileSet
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet


def from_file_to_json(filepath: epath.PathLike) -> tuple[epath.Path, Json]:
    """Loads the file as a JSON-LD.

    Args:
        filepath: The path to the file as a str or a path. The path can be absolute or
            relative.

    Returns:
        A tuple with the absolute path to the folder and the JSON-LD.
    """
    filepath = epath.Path(filepath).expanduser().resolve()
    if not filepath.exists():
        raise ValueError(f"File {filepath} does not exist.")
    with filepath.open() as f:
        data = json.load(f)
    return filepath.parent, data


def from_nodes_to_graph(metadata) -> nx.MultiDiGraph:
    """Converts the metadata to a structure graph linking nodes to their sources."""
    graph = nx.MultiDiGraph()
    # Bind graph to nodes:
    for node in metadata.nodes():
        graph.add_node(node)
    uuid_to_node = _check_no_duplicate(metadata)
    for node in metadata.distribution:
        if node.contained_in:
            for uuid in node.contained_in:
                _add_edge(graph, uuid_to_node, uuid, node)
    for record_set in metadata.record_sets:
        for field in record_set.fields:
            _add_edge(graph, uuid_to_node, record_set.uuid, field)
            for origin in [field.source, field.references]:
                if origin:
                    _add_edge(graph, uuid_to_node, origin.uuid, record_set)
            for sub_field in field.sub_fields:
                for origin in [sub_field.source, sub_field.references]:
                    if origin:
                        _add_edge(graph, uuid_to_node, origin.uuid, record_set)
    # `Metadata` are used as the entry node.
    _add_node_as_entry_node(graph, metadata)
    return graph


def _get_entry_nodes(graph: nx.MultiDiGraph, node: Node) -> list[Node]:
    """Retrieves the entry nodes (without predecessors) in a graph."""
    ctx = node.ctx
    entry_nodes: list[Node] = []
    for node, indegree in graph.in_degree(graph.nodes()):
        if indegree == 0:
            entry_nodes.append(node)  # pytype: disable=container-type-mismatch
    # Fields should usually not be entry nodes, except if they have subFields. So we
    # check for this:
    for node in entry_nodes:
        if isinstance(node, RecordSet) and not node.data:
            for field in node.fields:
                if not field.source:
                    field.add_error(
                        f'Node "{field.uuid}" is a field and has no source. Please, use'
                        f" {constants.ML_COMMONS_SOURCE(ctx)} to specify the source."
                    )
                else:
                    field.add_error(
                        f"Malformed source data: {field.source.uuid}. It does not refer"
                        " to any existing node. Have you used"
                        f" {constants.ML_COMMONS_FIELD(ctx)} or"
                        f" {constants.SCHEMA_ORG_DISTRIBUTION} to indicate the source"
                        " field or the source distribution? If you specified a field,"
                        " it should contain all the names from the RecordSet separated"
                        ' by `/`, e.g.: "record_set_name/field_name"'
                    )
    return entry_nodes  # pytype: disable=bad-return-type


def _check_no_duplicate(metadata) -> dict[str, Node]:
    """Checks no node has duplicated UUID and returns the mapping `uuid`->`Node`."""
    uuid_to_node: dict[str, Node] = {}
    for node in metadata.nodes():
        if node.uuid in uuid_to_node:
            node.add_error(
                "Duplicate nodes with the same identifier:"
                f" {uuid_to_node[node.uuid].uuid}"
            )
        uuid_to_node[node.uuid] = node
    return uuid_to_node


def _add_node_as_entry_node(graph: nx.MultiDiGraph, node: Node):
    """Add `node` as the entry node of the graph by updating `graph` in place."""
    graph.add_node(node, parent=None)
    entry_nodes = _get_entry_nodes(graph, node)
    for entry_node in entry_nodes:
        if isinstance(node, (FileObject, FileSet)):
            graph.add_edge(entry_node, node)


def _add_edge(
    graph: nx.MultiDiGraph,
    uuid_to_node: dict[str, Node],
    uuid: str,
    node: Node,
):
    """Adds an edge in the structure graph."""
    if uuid not in uuid_to_node:
        node.add_error(
            f'There is a reference to node with UUID "{uuid}" in node "{node.uuid}",'
            " but this node doesn't exist."
        )
        return
    graph.add_edge(uuid_to_node[uuid], node)



"""Join operation module."""

import dataclasses

import pandas as pd

from mlcroissant._src.operation_graph.base_operation import Operation
from mlcroissant._src.operation_graph.operations.field import apply_transforms_fn
from mlcroissant._src.structure_graph.nodes.field import Field
from mlcroissant._src.structure_graph.nodes.record_set import get_parent_uuid
from mlcroissant._src.structure_graph.nodes.record_set import RecordSet
from mlcroissant._src.structure_graph.nodes.source import Source


@dataclasses.dataclass(frozen=True, repr=False)
class Join(Operation):
    """Joins pd.DataFrames."""

    node: RecordSet

    def call(self, *args: pd.Series) -> pd.Series | None:
        """See class' docstring."""
        if len(args) == 1:
            return args[0]
        predecessors: list[str | None] = [
            operation.node.uuid for operation in self.operations.predecessors(self)
        ]
        if len(predecessors) != len(args):
            raise ValueError(f"Unsupported: Trying to join {len(args)} pd.DataFrames.")
        fields = self.node.fields
        # `joins` is the list of joins: field x (source1, df1) x (source2, df2)
        joins: list[
            tuple[Field, tuple[Source, pd.Series], tuple[Source, pd.Series]]
        ] = []
        for field in fields:
            left = field.source
            right = field.references
            if left is None or right is None:
                continue
            if left.uuid is None or right.uuid is None:
                continue
            left_index = predecessors.index(get_parent_uuid(self.node.ctx, left.uuid))
            right_index = predecessors.index(get_parent_uuid(self.node.ctx, right.uuid))
            join = (field, (left, args[left_index]), (right, args[right_index]))
            if join not in joins:
                joins.append(join)
        for field, (left, df_left), (right, df_right) in joins:
            assert left is not None and left.uuid is not None, (
                f'Left reference for "{field.uuid}" is None. It should be a valid'
                " reference."
            )
            assert right is not None and right.uuid is not None, (
                f'Right reference for "{field.uuid}" is None. It should be a valid'
                " reference."
            )
            left_column = left.get_column()
            right_column = right.get_column()
            assert left_column in df_left.columns, (
                f'Column "{left_column}" does not exist in node "{left.uuid}".'
                f" Existing columns: {df_left.columns}"
            )
            assert right_column in df_right.columns, (
                f'Column "{right_column}" does not exist in node "{right.uuid}".'
                f" Existing columns: {df_right.columns}"
            )
            df_left[left_column] = df_left[left_column].transform(
                apply_transforms_fn, field=field
            )
            return df_left.merge(
                df_right,
                left_on=left_column,
                right_on=right_column,
                how="left",
                suffixes=(None, "_right"),
            )
        return None



import pytest

from mlcroissant._src.core.context import CroissantVersion

VERSIONS = ["0.8", "1.0"]
CONFORMS_TO = [
    CroissantVersion(f"http://mlcommons.org/croissant/{version}")
    for version in VERSIONS
]

parametrize_version = pytest.mark.parametrize("version", VERSIONS)
parametrize_conforms_to = pytest.mark.parametrize(
    "conforms_to",
    CONFORMS_TO,
)



"""RecordSet module."""

import itertools
import json

from rdflib import term
from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.data_types import data_types_from_jsonld
from mlcroissant._src.core.data_types import data_types_to_jsonld
from mlcroissant._src.core.types import Json
from mlcroissant._src.core.uuid import formatted_uuid_to_json
from mlcroissant._src.core.uuid import uuid_from_jsonld
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.base_node import node_by_uuid
from mlcroissant._src.structure_graph.nodes.field import Field


def json_from_jsonld(ctx: Context, data) -> Json | None:
    """Creates `data` from a JSON-LD fragment."""
    if isinstance(data, str):
        try:
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            ctx.issues.add_error(
                f"{constants.ML_COMMONS_DATA(ctx)} is not a proper list of JSON: {data}"
            )
    return None


@mlc_dataclasses.dataclass
class RecordSet(Node):
    """Nodes to describe a dataset RecordSet."""

    JSONLD_TYPE = constants.ML_COMMONS_RECORD_SET_TYPE

    data: list[Json] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description="One or more records that constitute the data of the `RecordSet`.",
        from_jsonld=json_from_jsonld,
        url=constants.ML_COMMONS_DATA,
    )
    data_types: list[term.URIRef] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "The data type of the RecordSet. Mainly used to specify: `sc:Enumeration`."
        ),
        from_jsonld=data_types_from_jsonld,
        to_jsonld=data_types_to_jsonld,
        url=constants.ML_COMMONS_DATA_TYPE,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_DESCRIPTION,
    )
    examples: list[Json] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "One or more records provided as example content of the `RecordSet`, or a"
            " reference to data source that contains examples."
        ),
        from_jsonld=json_from_jsonld,
        url=constants.ML_COMMONS_EXAMPLES,
    )
    is_enumeration: bool | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Boolean],
        url=constants.ML_COMMONS_IS_ENUMERATION,
    )
    key: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "One or more fields whose values uniquely identify each record in the"
            " `RecordSet`."
        ),
        from_jsonld=lambda ctx, jsonld: uuid_from_jsonld(jsonld),
        to_jsonld=formatted_uuid_to_json,
        url=constants.SCHEMA_ORG_KEY,
    )
    name: str = mlc_dataclasses.jsonld_field(
        default="",
        description="The name of the RecordSet.",
        input_types=[SDO.Text],
        url=constants.SCHEMA_ORG_NAME,
    )
    fields: list[Field] = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "A data element that appears in the records of the RecordSet (e.g., one"
            " column of a table)."
        ),
        input_types=[Field],
        url=constants.ML_COMMONS_FIELD,
    )

    def __post_init__(self):
        """Checks arguments of the node."""
        Node.__post_init__(self)
        uuid_field = "name" if self.ctx.is_v0() else "id"
        self.validate_name()
        self.assert_has_mandatory_properties(uuid_field)

        if self.data is not None:
            data = self.data
            if not isinstance(data, list):
                self.add_error(
                    f"{constants.ML_COMMONS_DATA(self.ctx)} should declare a list. Got:"
                    f" {type(data)}."
                )
                return
            if not data:
                self.add_error(
                    f"{constants.ML_COMMONS_DATA(self.ctx)} should declare a non empty"
                    " list."
                )
            if self.ctx.is_v0():
                expected_keys = {field.name for field in self.fields}
            else:
                expected_keys = {field.id for field in self.fields}
            for i, line in enumerate(data):
                if not isinstance(line, dict):
                    self.add_error(
                        f"{constants.ML_COMMONS_DATA(self.ctx)} should declare a list"
                        f" of dict. Got: a list of {type(line)}."
                    )
                    return
                keys = set(line.keys())
                if keys != expected_keys:
                    self.add_error(
                        f"Line #{i} doesn't have the expected columns. Expected:"
                        f" {expected_keys}. Got: {keys}."
                    )

    def check_joins_in_fields(self):
        """Checks that all joins are declared when they are consumed."""
        joins: set[tuple[str, str]] = set()
        sources: set[str | None] = set()
        for field in self.fields:
            source_uuid = field.source.uuid
            references_uuid = field.references.uuid
            if source_uuid:
                # source_uuid is used as a source.
                sources.add(get_parent_uuid(self.ctx, source_uuid))
            if source_uuid and references_uuid:
                # A join happens because the user specified `source` and `references`.
                joins.add(
                    (
                        get_parent_uuid(self.ctx, source_uuid),
                        get_parent_uuid(self.ctx, references_uuid),
                    )
                )
                joins.add(
                    (
                        get_parent_uuid(self.ctx, references_uuid),
                        get_parent_uuid(self.ctx, source_uuid),
                    )
                )
        for combination in itertools.combinations(sources, 2):
            if combination not in joins:
                # Sort for reproducibility.
                ordered_combination = tuple(sorted(combination))
                self.add_error(
                    f"You try to use the sources with names {ordered_combination} as"
                    " sources, but you didn't declare a join between them. Use"
                    " `ml:references` to declare a join. Please, refer to the"
                    " documentation for more information."
                )


def get_parent_uuid(ctx: Context, uuid: str) -> str | None:
    """Retrieves the UID of the parent, e.g. `file/column` -> `file`."""
    node = node_by_uuid(ctx, uuid)
    if node is None:
        ctx.issues.add_error(
            f"Node with uuid={uuid} does not exist. This error might have been found"
            " during a Join operation."
        )
        return None
    if isinstance(node, Field):
        return node.parent.uuid
    return node.uuid



"""filter_test module."""

import pytest

from mlcroissant._src.operation_graph.operations.filter import FilterFiles
from mlcroissant._src.operation_graph.operations.filter import match_path
from mlcroissant._src.tests.nodes import empty_file_set
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = FilterFiles(operations=operations(), node=empty_file_set)
    assert str(operation) == "FilterFiles(file_set_name)"


@pytest.mark.parametrize(
    ["patterns", "path", "expected"],
    [
        [["**/*.csv"], "/path/to/this.csv", True],
        [["**/*.csv", "**/*.json"], "/path/to/this.csv", True],
        [["**/*.csv", "**/*.json"], "/path/to/this.parquet", False],
        [None, "/path/to/this.csv", True],
        [[], "/path/to/this.csv", True],
    ],
)
def test_match_path(patterns, path, expected):
    assert match_path(patterns, path) == expected



"""concatenate_test module."""

from mlcroissant._src.operation_graph.operations import concatenate
from mlcroissant._src.tests.nodes import empty_file_set
from mlcroissant._src.tests.operations import operations


def test_str_representation():
    operation = concatenate.Concatenate(operations=operations(), node=empty_file_set)
    assert str(operation) == "Concatenate(file_set_name)"



"""dataclasses_test module."""

import pytest
from rdflib.namespace import SDO

from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.structure_graph.base_node import Node


def test_dataclass():
    url = lambda ctx: "http://foo.org" if ctx.is_v0() else "http://bar.org"

    @mlc_dataclasses.dataclass
    class SomeNode:
        field1: int = mlc_dataclasses.jsonld_field(
            description="The first field",
            input_types=[SDO.Integer],
        )
        field2: list[str] = mlc_dataclasses.jsonld_field(
            cardinality="MANY",
            description="The second field",
            input_types=[SDO.Text],
            url=url,
        )

    node = SomeNode(field1=42, field2=["foo"])
    for cls_or_instance in [node, SomeNode]:
        field1, field2 = list(mlc_dataclasses.jsonld_fields(cls_or_instance))

        # Field #1
        assert field1.name == "field1"
        assert field1.cardinality == "ONE"
        assert field1.description == "The first field"
        assert field1.from_jsonld == None
        assert field1.input_types == [SDO.Integer]
        assert field1.to_jsonld == None
        assert field1.url == None

        # Field #2
        assert field2.name == "field2"
        assert field2.cardinality == "MANY"
        assert field2.description == "The second field"
        assert field2.from_jsonld == None
        assert field2.input_types == [SDO.Text]
        assert field2.to_jsonld == None
        assert field2.url == url
        assert (
            field2.call_url(Context(conforms_to=CroissantVersion.V_0_8))
            == "http://foo.org"
        )
        assert (
            field2.call_url(Context(conforms_to=CroissantVersion.V_1_0))
            == "http://bar.org"
        )


def test_jsonld_fields():
    @mlc_dataclasses.dataclass
    class SomeNode(Node):
        field1: int = mlc_dataclasses.jsonld_field(
            default=0,
            description="The first field",
            input_types=[SDO.Integer],
            versions=[CroissantVersion.V_0_8],
        )
        field2: str = mlc_dataclasses.jsonld_field(
            default="",
            description="The second field",
            input_types=[SDO.Text],
        )

    node = SomeNode(ctx=Context(conforms_to=CroissantVersion.V_1_0))
    fields = list(mlc_dataclasses.jsonld_fields(node))
    assert len(fields) == 1
    assert fields[0].name == "field2"


@pytest.mark.parametrize(
    "accessor",
    [
        # Access class
        lambda cls: cls,
        # Access instance of class
        lambda cls: cls(),
    ],
)
def test_types_of_fields(accessor):
    @mlc_dataclasses.dataclass
    class SubNode(Node):
        field: int | None = mlc_dataclasses.jsonld_field(
            default=None,
            input_types=[SDO.Text],  # this is incompatible with `int`
        )

    with pytest.raises(TypeError, match='Field "SubNode.field" should have type'):
        list(mlc_dataclasses.jsonld_fields(accessor(SubNode)))

    @mlc_dataclasses.dataclass
    class SubNode(Node):
        field: int = mlc_dataclasses.jsonld_field(
            default=None,  # this is incompatible with `int` (expect int | None)
            input_types=[SDO.Integer],
        )

    with pytest.raises(TypeError, match='Field "SubNode.field" should have type'):
        list(mlc_dataclasses.jsonld_fields(accessor(SubNode)))

    @mlc_dataclasses.dataclass
    class SubNode(Node):
        field: str = mlc_dataclasses.jsonld_field(
            default_factory=Node,  # this is incompatible with str
            input_types=[Node],
        )

    with pytest.raises(TypeError, match='Field "SubNode.field" should have type'):
        list(mlc_dataclasses.jsonld_fields(accessor(SubNode)))

    def cast_fn(value: Node) -> int:  # int is incompatible with Node
        del value
        return 0

    @mlc_dataclasses.dataclass
    class SubNode(Node):
        field: Node = mlc_dataclasses.jsonld_field(
            cast_fn=cast_fn,
            default_factory=Node,
            input_types=[Node],
        )

    with pytest.raises(TypeError, match='Field "SubNode.field" .* should have type'):
        list(mlc_dataclasses.jsonld_fields(accessor(SubNode)))



"""Utils to overload Python built-in dataclasses.

This module implements https://peps.python.org/pep-0681.
"""

from __future__ import annotations

from collections.abc import Iterator, Mapping, Sequence
import dataclasses
import inspect
from typing import Any, Callable, cast, Literal, TypedDict, Union

from rdflib import term
from rdflib.namespace import SDO
from typing_extensions import dataclass_transform

from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion
from mlcroissant._src.core.types import Json

MATCHING_TYPES: Mapping[term.URIRef, type] = {
    SDO.Boolean: bool,
    SDO.Date: str,
    SDO.DateTime: str,
    SDO.Integer: int,
    SDO.Language: str,
    SDO.Number: float,
    SDO.Text: str,
    SDO.Time: str,
    SDO.URL: str,
}


class Metadata(TypedDict):
    """Interface for the metadata mapping.

    Arguments:
        cardinality: The RDF cardinality ("ONE" or "MANY").
        cast_fn: A function to apply in the __init__ that will cast the input value to
            a value with the expected type by the dataclass.
        description: The description of the field.
        exclusive_with: List of all fields the current field should be exclusive with,
            i.e. they cannot be both set at the same time.
        from_jsonld: A function to overwrite the parsing JSON-LD->Python.
        to_jsonld: A function to overwrite the parsing Python->JSON-LD.
        required: Whether the field is required.
        url: The RDF URL for the field or a function that takes into input the context
            (ctx) and outputs the RDF URL for the field.
        versions: The list/set of Croissant versions for which this field is used (by
            default all versions).
    """

    cardinality: Literal["ONE", "MANY"]
    cast_fn: Callable[[Any], Any] | None
    description: str
    exclusive_with: Sequence[str]
    from_jsonld: Callable[[Context, Json], Any] | None
    input_types: list[term.URIRef | type]
    to_jsonld: Callable[[Context, Any], Json] | None
    url: term.URIRef | Callable[[Context], term.URIRef]
    versions: set[CroissantVersion] | list[CroissantVersion]


@dataclasses.dataclass
class JsonldField:
    """Overloads dataclasses.Field with JSON-LD-specific attributes."""

    default: Any
    default_factory: Any
    name: str
    metadata: Metadata

    @property
    def cardinality(self):
        """Getter for cardinality."""
        return self.metadata.get("cardinality")

    @property
    def cast_fn(self):
        """Getter for cast_fn."""
        return self.metadata.get("cast_fn")

    @property
    def description(self):
        """Getter for description."""
        return self.metadata.get("description")

    @property
    def exclusive_with(self):
        """Getter for exclusive_with."""
        return self.metadata.get("exclusive_with")

    @property
    def from_jsonld(self):
        """Getter for from_jsonld."""
        return self.metadata.get("from_jsonld")

    @property
    def input_types(self):
        """Getter for input_types."""
        return self.metadata.get("input_types", [])

    @property
    def to_jsonld(self):
        """Getter for to_jsonld."""
        return self.metadata.get("to_jsonld")

    @property
    def url(self):
        """Getter for url."""
        return self.metadata.get("url")

    @property
    def versions(self):
        """Getter for versions."""
        return self.metadata.get("versions")

    def call_from_jsonld(self, ctx: Context, value: Any):
        """Calls `from_jsonld` in the field."""
        if value and self.from_jsonld:
            return self.from_jsonld(ctx, value)
        else:
            return value

    def call_to_jsonld(self, ctx: Context, value: Any):
        """Calls `to_jsonld` in the field."""
        if value and self.to_jsonld:
            return self.to_jsonld(ctx, value)
        else:
            return value

    def call_url(self, ctx: Context) -> term.URIRef:
        """Calls `jsonld` in the field."""
        url = self.url
        if callable(url):
            return url(ctx)
        else:
            return url


def jsonld_field(
    cardinality="ONE",
    cast_fn=None,
    description="",
    exclusive_with=None,
    from_jsonld=None,
    input_types=None,
    to_jsonld=None,
    url=None,
    versions=None,
    **kwargs,
):
    """Overloads dataclasses.field with specific attributes."""
    if cardinality not in ["ONE", "MANY"]:
        raise ValueError(f"cardinality should be ONE or MANY. Got {cardinality}")
    if input_types is None:
        input_types = []
    if exclusive_with is None:
        exclusive_with = []
    if versions is None:
        versions = set(CroissantVersion)  # All versions are supported by default
    if not from_jsonld and not input_types:
        raise ValueError("You should specify either `from_jsonld` or `input_types`.")
    elif from_jsonld and input_types:
        raise ValueError("You can only specify either `from_jsonld` or `input_types`.")
    return dataclasses.field(
        metadata=Metadata(
            cardinality=cardinality,
            cast_fn=cast_fn,
            description=description,
            exclusive_with=exclusive_with,
            from_jsonld=from_jsonld,
            input_types=input_types,
            to_jsonld=to_jsonld,
            url=url,
            versions=versions,
        ),
        **kwargs,
    )


def jsonld_fields(cls_or_instance) -> Iterator[JsonldField]:
    """Filters the JSON-LD fields."""
    for field in dataclasses.fields(cls_or_instance):
        if hasattr(cls_or_instance, "ctx"):
            ctx = getattr(cls_or_instance, "ctx")
            versions = field.metadata.get("versions") or set()
            if isinstance(ctx, Context) and ctx.conforms_to not in versions:
                # The field is not to be used with the current version of Croissant:
                continue
        metadata = cast(Metadata, field.metadata)
        if metadata:
            _check_types(cls_or_instance, field, metadata)
            yield JsonldField(
                default=field.default,
                default_factory=field.default_factory,
                name=field.name,
                metadata=metadata,
            )


@dataclass_transform(
    field_specifiers=(jsonld_field,)
)  # pytype: disable=not-supported-yet
def dataclass(cls):
    """Overloads the built-in dataclass with JsonldFields instead of Fields."""
    return dataclasses.dataclass(eq=False, repr=False)(cls)


def _check_types(cls_or_instance, field: dataclasses.Field, metadata: Metadata) -> None:
    """Checks that the JSON-LD types confirm to the Python types for cls_or_instance."""
    metadata = Metadata(**metadata)
    input_types = metadata["input_types"]
    cls_name = (
        cls_or_instance.__name__
        if isinstance(cls_or_instance, type)
        else cls_or_instance.__class__.__name__
    )
    field_name = f'"{cls_name}.{field.name}"'
    if not input_types:
        # No heuristic is possible as input_types were not specified:
        return
    types: list[type] = []
    for input_type in input_types:
        if isinstance(input_type, term.URIRef):
            if input_type in MATCHING_TYPES:
                types.append(MATCHING_TYPES[input_type])
            else:
                raise ValueError(f"input_type={input_type} has no matching type.")
        else:
            types.append(input_type)
    expected_type = Union[tuple(types)]  # type: ignore
    if metadata["cardinality"] == "MANY":
        expected_type = list[expected_type]  # type: ignore
    if field.default != dataclasses.MISSING:
        expected_type = Union[expected_type, type(field.default)]  # type: ignore

    cast_fn = metadata["cast_fn"]
    if cast_fn:
        # The field defines cast_fn: we check that Python type == cast_fn output type.
        signature = inspect.signature(cast_fn)
        name = f"{field_name} with cast_fn={cast_fn.__name__}"
        _types_are_equal(name, field.type, signature.return_annotation)
    else:
        # Python type == expected JSON-LD type.
        _types_are_equal(field_name, field.type, expected_type)


def _types_are_equal(field_name: str, type1: Any, type2: Any) -> None:
    if type1 != type2:
        raise TypeError(f"Field {field_name} should have type {type2}. Got {type1}")



"""Init operation module."""

import dataclasses
import logging

from mlcroissant._src.operation_graph.base_operation import Operation


@dataclasses.dataclass(frozen=True, repr=False)
class InitOperation(Operation):
    """Sets up other operations."""

    def call(self, *args):
        """See class' docstring."""
        del args  # unused
        logging.info("Setting up generation for dataset: %s", self.node.uuid)



"""Beam module."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from etils import epath

from mlcroissant._src.datasets import Dataset
from mlcroissant._src.datasets import Filters


def ReadFromCroissant(
    *,
    jsonld: epath.PathLike | Mapping[str, Any],
    record_set: str,
    mapping: Mapping[str, epath.PathLike] | None = None,
    filters: Filters | None = None,
):
    """Returns an Apache Beam PCollection to generate the dataset using e.g. Spark.

    Example of usage:

    ```python
    import apache_beam as beam
    from apache_beam.options import pipeline_options
    import mlcroissant as mlc

    jsonld = "https://huggingface.co/api/datasets/ylecun/mnist/croissant"

    options = pipeline_options.PipelineOptions()
    with beam.Pipeline(options=options) as pipeline:
        _ = pipeline | mlc.ReadFromCroissant(
            jsonld=jsonld,
            record_set="mnist",
        )
    ```

    The sharding is done on the filtered files. This is currently optimized for Hugging
    Face datasets, so it raises an error if the dataset is not a Hugging Face dataset.

    Args:
        jsonld: A JSON object or a path to a Croissant file (URL, str or pathlib.Path).
        record_set: The name of the record set to generate.
        mapping: Mapping filename->filepath as a Python dict[str, str] to handle manual
            downloads. If `document.csv` is the FileObject and you downloaded it to
            `~/Downloads/document.csv`, you can specify `mapping={"document.csv":
            "~/Downloads/document.csv"}`.
        filters: A dictionary mapping a field ID to the value we want to filter in. For
            example, when writing {'data/split': 'train'}, we want to keep all records
            whose field `data/split` takes the value `train`.

    Returns:
        A Beam PCollection with all the records where each element contains a tuple with
            a) a global index, and
            b) the content of the record.

    Raises:
        A ValueError if the dataset is not streamable.
    """
    dataset = Dataset(jsonld=jsonld, mapping=mapping)
    return dataset.records(record_set, filters=filters).beam_reader()



"""FileObject module."""

from rdflib.namespace import SDO

from mlcroissant._src.core import constants
from mlcroissant._src.core import dataclasses as mlc_dataclasses
from mlcroissant._src.core.constants import ML_COMMONS
from mlcroissant._src.core.uuid import formatted_uuid_to_json
from mlcroissant._src.core.uuid import uuid_from_jsonld
from mlcroissant._src.structure_graph.base_node import Node
from mlcroissant._src.structure_graph.nodes.source import Source


@mlc_dataclasses.dataclass
class FileObject(Node):
    """Nodes to describe a dataset FileObject (distribution)."""

    content_url: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "Actual bytes of the media object, for example the image file or video"
            " file."
        ),
        input_types=[SDO.URL],
        url=SDO.contentUrl,
    )
    content_size: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description=(
            "File size in (mega/kilo/...)bytes. Defaults to bytes if a unit is not"
            " specified."
        ),
        input_types=[SDO.Text],
        url=SDO.contentSize,
    )
    contained_in: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default_factory=list,
        description=(
            "Another FileObject or FileSet that this one is contained in, e.g., in the"
            " case of a file extracted from an archive. When this property is present,"
            " the contentUrl is evaluated as a relative path within the container"
            " object"
        ),
        from_jsonld=lambda ctx, contained_in: uuid_from_jsonld(contained_in),
        to_jsonld=lambda ctx, contained_in: [
            formatted_uuid_to_json(ctx, uuid) for uuid in contained_in
        ],
        url=SDO.containedIn,
    )
    description: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=SDO.description,
    )
    encoding_format: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        description="The format of the file, given as a mime type.",
        input_types=[SDO.Text],
        url=SDO.encodingFormat,
    )
    md5: str | None = mlc_dataclasses.jsonld_field(
        default=None,
        input_types=[SDO.Text],
        url=lambda ctx: ML_COMMONS(ctx).md5,
    )
    name: str = mlc_dataclasses.jsonld_field(
        default="",
        description=(
            "The name of the file.  As much as possible, the name should reflect the"
            " name of the file as downloaded, including the file extension. e.g."
            ' "images.zip".'
        ),
        input_types=[SDO.Text],
        url=SDO.name,
    )
    same_as: list[str] | None = mlc_dataclasses.jsonld_field(
        cardinality="MANY",
        default=None,
        description=(
            "URL (or local name) of a FileObject with the same content, but in a"
            " different format."
        ),
        input_types=[SDO.URL],
        url=SDO.sameAs,
    )
    sha256: str | None = mlc_dataclasses.jsonld_field(
        cardinality="ONE",
        default=None,
        description="Checksum for the file contents.",
        input_types=[SDO.Text],
        url=SDO.sha256,
    )
    source: Source = mlc_dataclasses.jsonld_field(
        default_factory=Source,
        input_types=[Source],
        url=lambda ctx: ML_COMMONS(ctx).source,
    )

    def __post_init__(self):
        """Checks arguments of the node."""
        Node.__post_init__(self)
        self.validate_name()
        uuid_field = "name" if self.ctx.is_v0() else "id"
        self.assert_has_mandatory_properties("encoding_format", uuid_field)

        if not self.contained_in:
            self.assert_has_mandatory_properties("content_url")
            if not self.ctx.is_live_dataset:
                self.assert_has_exclusive_properties(["md5", "sha256"])

    JSONLD_TYPE = constants.SCHEMA_ORG_FILE_OBJECT



"""Tests for CroissantVersion."""

import pytest

from mlcroissant._src.core.context import Context
from mlcroissant._src.core.context import CroissantVersion


@pytest.mark.parametrize(
    "conforms_to",
    [1, 1.0, "1.0"],
)
def test_conforms_to_is_invalid(conforms_to):
    ctx = Context(conforms_to=conforms_to)
    assert any(
        error.startswith("conformsTo should be a string or a CroissantVersion.")
        for error in ctx.issues.errors
    )


@pytest.mark.parametrize(
    ["conforms_to", "expected"],
    [
        [None, CroissantVersion.V_0_8],
        ["http://mlcommons.org/croissant/0.8", CroissantVersion.V_0_8],
        ["http://mlcommons.org/croissant/1.0", CroissantVersion.V_1_0],
        [CroissantVersion.V_0_8, CroissantVersion.V_0_8],
        [CroissantVersion.V_1_0, CroissantVersion.V_1_0],
    ],
)
def test_conforms_to_is_checked(conforms_to, expected: CroissantVersion):
    # If left empty, conforms_to defaults to 0.8.
    ctx = Context()
    assert CroissantVersion.from_jsonld(ctx, conforms_to) == expected
    assert not ctx.issues.errors
    assert not ctx.issues.warnings
